/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.datamodel;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.util.io.LazyInputStream;
import org.exoplatform.services.log.ExoLogger;

/**
 * AbstractValueData class provide common functions for value data storing. 
 * With support of temporary swapping in local files.
 * AbstractValueData can store value data as array and as stream of bytes.
 * 
 * In most of cases data will be stored as array of bytes.
 * In case of BLOB (binary large objects) it may be stored in input stream.
 * 
 * If used 'inBuffer' option (available with InputStream instance) then given stream will be 
 * copied into bytes array.   
 * Otherwise, a ValueData will store given InputStream as direct source of data access in
 * getAsStream() method.
 * 
 * If we want to use local swapping inside value data object we have to use constructors with 'cachePath' parameter.
 * This is a string contains file path will be used for local swapping of data.
 * But swapping will be available for BLOB (binary large objects) values only.
 * Data will be detected as BLOB if data has size (bytes count) bigger given limit (MAX_BUFFER_SIZE).
 * 
 * If swap is enabled a BLOB file will be stored as local file, nor bytes/stream will be stored in class instance.
 * E.g. if we call AbstractValueData(InputStream, int, String) on BLOB we will store the stream content in a local file.
 * NOTE: no stream object will be stored in value data. 
 * But for non-BLOB values constructor AbstractValueData(InputStream, int, String) 
 * will store data as array of bytes in memory. Nor file, nor stream will be created/stored here.     
 * 
 * Created by The eXo Platform SARL        .
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: AbstractValueData.java 8661 2006-09-07 10:17:32Z peterit $
 */

public abstract class AbstractValueData implements ValueData {
    
  protected static final int BLOB_BUFFER_SIZE = 200 * 1024; // 200Kb default binary values size
  
  protected static final int BLOB_MIN_SIZE = 2048 * 1024; // 2Mb default BLOB file size

  private static Log log = ExoLogger.getLogger("jcr.AbstractValueData");
  
  protected final InputStream stream;
  protected final byte[] data;
  
  protected final File swapFile; 

  protected int orderNumber; 

  /**
   * Create value data storing data as bytes array
   * 
   * @param data - bytes
   * @param orderNumber - value order nymber
   */
  protected AbstractValueData(final byte[] data, int orderNumber) {
    this.data = data;
    this.stream = null;
    
    this.swapFile = null;
    this.orderNumber = orderNumber;
  }
  
  // No use case can be found
//  protected AbstractValueData(final byte[] data, final int orderNumber, String cachePath) throws IOException {
//    this.data = data;
//    this.stream = null;
//    
//    this.orderNumber = orderNumber;
//    
//    File cFile = new File(cachePath);
//    FileOutputStream out = new FileOutputStream(cFile);
//    FileChannel cacheChannel = out.getChannel();
//    cacheChannel.write(ByteBuffer.wrap(data));
//    
//    this.cacheFile = cFile;
//    this.enableCache = (cFile != null);
//  }
  
  /**
   * Create value data storing data as lazy readable input stream (given as external)
   * 
   * @param stream - LazyInputStream instance 
   * @param orderNumber - value order nymber
   */
  protected AbstractValueData(final LazyInputStream stream, int orderNumber) throws IOException {
    this(stream, orderNumber, false); 
  }
  
  /**
   * Create value data storing data from input stream (given as external).
   * If 'inBuff' flag switched on then the stream will be stored as bytes array in memory.
   * Otherwise the stream will be used as direct source of data (with mark/reset abilities of a InputStream).
   * 
   * @param stream - InputStream instance 
   * @param orderNumber - value order nymber
   * @param inBuffer - if true, the stream will be stored as bytes array  
   */
  protected AbstractValueData(final InputStream stream, int orderNumber, boolean inBuffer) throws IOException {
    
    if (inBuffer) {
      // read given stream into bytes buffer
      this.stream = null;
      
      if (stream instanceof FileInputStream)
        this.data = readFile((FileInputStream) stream);
      else
        this.data = readStream(stream); 
    } else {
      // use given stream
      this.data = null;
      
      // We want to use mark/reset abilities of a InputStream
      // TODO But! It can use more memory in case of BLOB streams if use mark/reset methods.
      if (stream instanceof BufferedInputStream) // stream instanceof LazyInputStream || 
        this.stream = stream;
      else 
        this.stream = new BufferedInputStream(stream);
    }
    
    this.swapFile = null;
    
    this.orderNumber = orderNumber;
  }
  
  // No use case found - lazy readable and swap enabled.  
//  /**
//   * Create value data storing data as lazy readable input stream (given as external) with swap option enabled.
//   * I.e. if size of data stored in a stream more than maxBufferSize then this stream will be copied to swap file.
//   * And value data will use this swap file instead use of original stream.
//   * NOTE: stream will not stored in instance of ValueData.   
//   * 
//   * @param stream - LazyInputStream instance 
//   * @param orderNumber - value order nymber
//   * @param maxBufferSize - maximim size of memory buffer  
//   * @param swapFilePath - swap file path (probably local or mounted on local file system)
//   */
//  protected AbstractValueData(final LazyInputStream stream, int orderNumber, String swapFilePath, int maxBufferSize) throws IOException {
//    this((InputStream) stream, orderNumber, swapFilePath, maxBufferSize);
//  }
  
  /**
   * Create value data storing data as input stream (given as external) with swap option enabled.
   * I.e. if size of data stored in a stream more than maxBufferSize then this stream will be copied to swap file.
   * And value data will use this swap file instead use of original stream.
   * NOTE: stream will not stored in instance of ValueData.   
   * 
   * @param stream - InputStream instance 
   * @param orderNumber - value order nymber
   * @param maxBufferSize - maximim size of memory buffer  
   * @param swapFilePath - swap file path (probably local or mounted on local file system)
   */
  protected AbstractValueData(final InputStream stream, int orderNumber, String swapFilePath, int minSwapFileSize) throws IOException {
    byte[] buffer = new byte[minSwapFileSize + 64];
    int read;
    //int len = 0;
    FileOutputStream out = null;
    File swpFile = null;
    FileChannel swapChannel = null;
    ByteArrayOutputStream buff = null;
    // TODO have to make read cycle in independent thread with a result semaphore
    try {
      if (stream instanceof FileInputStream) {
        // direct bulk-read from file to file
        log.info("ValueData: Using direct bulk-read from file to file... " + swapFilePath);
        long start = System.currentTimeMillis();
        
        FileChannel sourceChannel = ((FileInputStream) stream).getChannel();
        
        swpFile = new File(swapFilePath);
        out = new FileOutputStream(swpFile);
        swapChannel = out.getChannel();
        swapChannel.transferFrom(sourceChannel, 0, sourceChannel.size());
        
        swapChannel.close();
        sourceChannel.close();
        
        log.info("ValueData: Direct bulk-read from file to file done. " 
            + (System.currentTimeMillis() - start) + "ms. " + swapFilePath);
      } else {
        // Reading stream to mapped swap file
        log.info("ValueData: Reading stream " + stream.getClass().getCanonicalName() + " to mapped swap file..." + swapFilePath);
        long start = System.currentTimeMillis();
        
        // try to mark stream
        if (stream.markSupported())
          try {
              stream.mark(stream.available());
          } catch (IOException e) {
            log.error("Can't mark BinaryValue stream " + stream + ". Error: " + e.getMessage(), e);
          }
        else
          log.warn("Can't mark BinaryValue stream. A stream will have no data for reading after (will be consumed). Re-read the value again. " + stream);
        
        int fileSize = 0;
        while ((read = stream.read(buffer)) > 0) {
          if (swapChannel != null) {
            // spool to swap file
            MappedByteBuffer swapBuffer = swapChannel.map(
                FileChannel.MapMode.READ_WRITE, 
                fileSize =+ read,  // swapChannel.size()
                read);
            swapBuffer.put(buffer, 0, read);
            
            //swapChannel.write(ByteBuffer.wrap(buffer, 0, read));
            
            //out.write(buffer, 0, read);
            //len += read;
          //} else if (len + read > MAX_BUFFER_SIZE) {
          } else if (buff == null && read > minSwapFileSize) {  
            // Threshold for keeping data in memory exceeded;
            // create swap file and spool buffer contents
            swpFile = new File(swapFilePath);
            out = new FileOutputStream(swpFile);
            swapChannel = out.getChannel();
            
            MappedByteBuffer swapBuffer = swapChannel.map(
                FileChannel.MapMode.READ_WRITE, 
                0, 
                fileSize = read);
            swapBuffer.put(buffer, 0, read);
            
            //swapChannel.write(ByteBuffer.wrap(buffer, 0, read));
            
  //          out.write(buffer, 0, len);
  //          out.write(buffer, 0, read);
            //data = null;
            //len += read;
          } else if (buff != null && buff.size() > minSwapFileSize) {
            // Threshold for keeping data in memory exceeded (after multiple reads);
            // create swap file and spool buffer contents
            swpFile = new File(swapFilePath);
            out = new FileOutputStream(swpFile);
            swapChannel = out.getChannel();
            
            MappedByteBuffer swapBuffer = swapChannel.map(
                FileChannel.MapMode.READ_WRITE, 
                0, 
                fileSize = buff.size() + read);
            swapBuffer.put(buff.toByteArray(), 0, buff.size());
            swapBuffer.put(buffer, 0, read);
            
            //swapChannel.write(ByteBuffer.wrap(buff.toByteArray()));
            //swapChannel.write(ByteBuffer.wrap(buffer, 0, read));
            
            //data = null;
          } else {
            // Storing data into temp bytes stream 
            if (buff == null)
              buff = new ByteArrayOutputStream();
            buff.write(buffer, 0, read);
            
            // reallocate new buffer and spool old buffer contents
  //          byte[] newBuffer = new byte[len + read];
  //          System.arraycopy(buffer, 0, newBuffer, 0, len);
  //          System.arraycopy(buffer, 0, newBuffer, len, read);
            //data = newBuffer;
            //len += read;
          }
        }
        log.info("ValueData: Stream " + stream.getClass().getCanonicalName() + " read to mapped swap file done. " 
            + (System.currentTimeMillis() - start) + "ms. " + swapFilePath);
      }
    
    } finally {
      if (swapChannel != null)
        swapChannel.close();
      if (out != null)
        out.close();
      if (stream.markSupported())
        try {
            stream.reset();
        } catch (IOException e) {
          log.error("Can't reset marked BinaryValue stream " + stream + ". Error " + e.getMessage(), e);
        }
    }
    
    if (swpFile == null)
      this.data = buff.toByteArray();
    else
      this.data = null;

    this.stream = null;
    
    this.swapFile = swpFile;
    
    this.orderNumber = orderNumber;
  }
 
  protected byte[] readStream(InputStream aStream) throws IOException {
    // optimizing small size data reading and memory usage till stream read operation
    int buffSize = aStream.available();
    byte[] buff = new byte[buffSize > 2048 ? (buffSize > BLOB_BUFFER_SIZE ? BLOB_BUFFER_SIZE : buffSize >>> 1) : 1024];

    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    int res = 0;
    while ((res = aStream.read(buff)) > 0) {
      baos.write(buff, 0, res);
    }
    return baos.toByteArray();
  }
  
  protected byte[] readFile(FileInputStream file) throws IOException {
    FileChannel channel = null;
    try {
      channel = file.getChannel();
      
      int size = (int) channel.size();
      
      if (size <= BLOB_MIN_SIZE) {
        // reading in classical way
        return readStream(file); 
      }
      
      // using memory-mapped file
      MappedByteBuffer fileBuffer = channel.map(FileChannel.MapMode.READ_ONLY, 0, size);
      
      byte[] buff = new byte[size];
      fileBuffer.get(buff);
      fileBuffer.rewind();
      
      return buff;
    } finally {
      if (channel != null)
        channel.close();
    }
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getOrderNumber()
   */
  public final int getOrderNumber() {
    return orderNumber;
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getAsByteArray()
   */
  public final byte[] getAsByteArray() throws IllegalStateException {
    if(data != null) {
      return data;
    } else if(swapFile != null) {
      throw new IllegalStateException("Data could not be rendered as byte array (large data)");
    } else {
      throw new NullPointerException("Data is null");
    }
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#isByteArray()
   */
  public final boolean isByteArray() {
    return data != null;
  }


  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#getAsStream()
   */
  public final InputStream getAsStream() throws IOException {
    // [PN] 07.09.06 We have to clone stream up from here
    if(stream != null) {
      if (stream instanceof LazyInputStream)
        return ((LazyInputStream) stream).clone();
      return stream;
    } else if(data != null)
      return new ByteArrayInputStream(data);
    else if(swapFile != null)
      return new FileInputStream(swapFile);
    else
      throw new NullPointerException("Data is null");
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.datamodel.ValueData#setOrderNumber(int)
   */
  public void setOrderNumber(int orderNumber) {
    this.orderNumber = orderNumber;
  }
  
  protected void finalize() throws Throwable {
    try {
      if(swapFile != null) { // && enableSwap
        if(!swapFile.delete())
          throw new IllegalStateException("Can not delete temporary file "+swapFile.getAbsolutePath());
      }
    } finally {
      super.finalize();
    }
  }
}
