/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.Item;
import javax.jcr.ItemNotFoundException;
import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.PropertyType;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.ItemDefinition;
import javax.jcr.nodetype.PropertyDefinition;
import javax.jcr.version.VersionException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.core.value.BaseValue;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: ItemImpl.java 8292 2006-08-26 17:37:27Z geaz $
 */

public abstract class ItemImpl implements Item {

  protected static Log log = ExoLogger.getLogger("jcr.ItemImpl");

  protected final SessionImpl session;

  /**
   * @deprecated
   */
  protected JCRPath location;
  
  protected ItemData data;

  protected SessionDataManager dataManager;

  protected LocationFactory locationFactory;

  protected ValueFactoryImpl valueFactory;

  protected int itemHashCode = 0; // [PN] 19.04.06

  ItemImpl(ItemData data, SessionImpl session) throws RepositoryException {

    this.session = session;
    this.data = data;
    this.location = session.getLocationFactory().createJCRPath(
        getData().getQPath());

    this.dataManager = session.getTransientNodesManager();
    this.locationFactory = session.getLocationFactory();
    this.valueFactory = (ValueFactoryImpl) session.getValueFactory();

    itemHashCode = (session.getWorkspace().getName() + data.getUUID())
        .hashCode();
  }

  // void loadLocation(JCRPath location) {
  // this.location = location;
  // }

  /*
   * (non-Javadoc)
   * 
   * @see javax.jcr.Item#getPath()
   */
  public String getPath() {
    return getLocation().getAsString(false);
  }

  /**
   * @see javax.jcr.Item#getName
   */
  public String getName() {
    return getLocation().getName().getAsString();
  }

  /**
   * @see javax.jcr.Item#getAncestor
   */
  public Item getAncestor(int degree) throws ItemNotFoundException,
      AccessDeniedException, RepositoryException {
    try {
      // 6.2.8 If depth > n is specified then an ItemNotFoundException is
      // thrown.
      if (degree < 0)
        throw new ItemNotFoundException(
            "Workspace.getAncestor() ancestor's degree < 0.");
      int n = getDepth() - degree;
      // log.debug("Item.getAncestor(" + degree + ") depth " +
      // getDepth()+" "+n);
      if (n == 0)
        return this;
      else if (n < 0)
        throw new ItemNotFoundException(
            "Workspace.getAncestor() ancestor's degree > depth of this item.");
      else {

        JCRPath ancestorLoc = getLocation().makeAncestorPath(n);

        return item(ancestorLoc);
      }
    } catch (PathNotFoundException e) {
      throw new ItemNotFoundException(e.getMessage(), e);
    }
  }

  /**
   * Parent is lazy initialized
   * 
   * @see javax.jcr.Item#getParent
   */
  public Node getParent() throws ItemNotFoundException, AccessDeniedException,
      RepositoryException {
    if (isRoot())
      throw new ItemNotFoundException("Root node does not have a parent");

    return parent();
  }

  /**
   * @see javax.jcr.Item#getSession
   */
  public Session getSession() {
    return session;
  }

  /**
   * @see javax.jcr.Item#getDepth
   */
  public int getDepth() {
    return getLocation().getDepth();
  }

  /**
   * @see javax.jcr.Item#isSame
   */
  public boolean isSame(Item otherItem) {
    if (otherItem == null)
      return false;

    if (!(otherItem instanceof ItemImpl))
      return false;

    try {
      // return this.getActualItemData().getUUID().equals(
      // ((ItemImpl) otherItem).getActualItemData().getUUID());

      // [PN] 18.04.06 UUID not changes on ItemImpl
      return getInternalUUID().equals(((ItemImpl) otherItem).getInternalUUID());
    } catch (Exception e) {
      log.debug("Item.isSame() failed " + e);
      return false;
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.jcr.Item#isNew()
   */
  public boolean isNew() {
    return dataManager.isNew(getInternalUUID());
  }

  /*
   * (non-Javadoc)
   * 
   * @see javax.jcr.Item#isModified()
   */
  public boolean isModified() {
    return dataManager.isModified(getInternalUUID());
  }

  /**
   * @see javax.jcr.Item#remove
   */
  public void remove() throws RepositoryException,
      ConstraintViolationException, VersionException, LockException {

    doRemove(true);
  }

  public void doRemove(boolean externalValidation) throws RepositoryException,
      ConstraintViolationException, VersionException, LockException {

    if (isRoot())
      throw new RepositoryException("Can't remove ROOT node.");

    // Check constraints
    if (externalValidation) {
      ItemDefinition def;
      if (isNode())
        def = ((NodeImpl) this).getDefinition();
      else
        def = ((PropertyImpl) this).getDefinition();

      if (def.isMandatory() || def.isProtected())
        throw new ConstraintViolationException(
            "Can't remove mandatory or protected item " + getPath());
    }
    // Check if versionable ancestor is not checked-in
    // parent = (NodeImpl) getParent();

    if (parent().isCheckedOut())
      throw new VersionException("Node " + parent().getPath()
          + " or its nearest ancestor is checked-in");

    // Check locking
    if (!parent().checkLocking())
      throw new LockException("Node " + parent().getPath() + " is locked ");


    // [1] remove from datamanager
    dataManager.delete(ItemState.createDeleted(data));

    // log.debug("REMOVED " + this.getState());
  }

  /**
   * @see javax.jcr.Item#save
   */
  public void save() throws ReferentialIntegrityException,
      AccessDeniedException, LockException, ConstraintViolationException,
      InvalidItemStateException, ReferentialIntegrityException,
      RepositoryException {

    if (isNew())
      throw new RepositoryException(
          "It is impossible to call save() on the newly added item "
              + getPath());
    dataManager.commit(getInternalPath());
  }

  /**
   * @see javax.jcr.Item#refresh
   */
  public void refresh(boolean keepChanges) throws InvalidItemStateException,
      RepositoryException {
    if (!keepChanges)
      dataManager.rollback(this.getData());
    else
      dataManager.refresh(this.getData());
  }

  // -------------------- Implementation ----------------------


  public String getInternalUUID() {
    return getData().getUUID();
  }

  public String getParentUUID() {
    return getData().getParentUUID();
  }

  public InternalQPath getInternalPath() {
    return getData().getQPath();
  }

  public InternalQName getInternalName() {
    return getData().getQPath().getName();
  }

  public JCRPath getLocation() {
    return location;
  }

  public ItemData getData() {
    return data;
  }

  public boolean isRoot() {
    return (getDepth() == 0);
  }

//  public int getState() {
//    return state.getState();
//  }

//  public ItemState getItemState() {
//    return state;
//  }

  public abstract void loadData(ItemData data) throws RepositoryException;

  public boolean hasPermission(String action) throws RepositoryException {

    // versionStorage's children should have READ permission for all
    if (isRoot()
        || (location.isDescendantOf(locationFactory
            .parseAbsPath("/jcr:system/jcr:versionStorage"), false) && action
            .equals(PermissionType.READ)))
      return true;

    // else if (parent == null) {
    // // [PN] 16.02.06 data used instead getActualItemData(), actualy it's
    // same.
    // // Cause - a cyclic method calls on not initialized node (no childs lazy
    // loaded)
    // parent = (NodeImpl) this.dataManager.getItemByUUID(data.getParentUUID());
    // }
    //
    // if (parent == null) {
    // parent = (NodeImpl) getParent();
    // }
    //    
    // if (parent == null) {
    // throw new RepositoryException("FATAL: parent not found for " +
    // this.getPath());
    // }

    // NodeImpl parentNode = (NodeImpl) ((isRoot()) ? (Node) this :
    // (Node)this.dataManager.getItemByUUID(getActualItemData().getParentUUID()));
    boolean hasPerm = session.getAccessManager().hasPermission(
        parent().getACL(), action, session.getUserID()); //.getChildrenACL()
    return hasPerm;
  }

  protected ItemImpl item(JCRPath path) throws RepositoryException {
    return dataManager.getItem(path.getInternalPath(), true);
  }
  
  protected ItemImpl item(InternalQPath path) throws RepositoryException {
    return dataManager.getItem(path, true);
  }

  protected ItemImpl item(String uuid) throws RepositoryException {
    return dataManager.getItemByUUID(uuid, true);
  }

  protected NodeImpl parent() throws RepositoryException {
    return (NodeImpl) item(getData().getParentUUID());
  }

  protected TransientPropertyData doUpdateProperty(NodeImpl parentNode,
      InternalQName propertyName, List<Value> propertyValues,
      boolean multiValue, int expectedType) throws ValueFormatException,
      VersionException, LockException, ConstraintViolationException,
      RepositoryException {

    PropertyDefinition def = parentNode
        .findChildPropertyDefinition(propertyName);

    if (def.isProtected())
      throw new ConstraintViolationException("Can not set protected property "
          + getPath());

    if (!parentNode.isCheckedOut())
      throw new VersionException("Node " + parent().getPath()
          + " or its nearest ancestor is checked-in");

    // Check locking
    if (!parentNode.checkLocking())
      throw new LockException("Node " + parent().getPath() + " is locked ");

    // no such definition
    if (def.isMultiple() != multiValue) {
      if (multiValue) {
        throw new ValueFormatException(
            "Can not assign multiple-values Value to a single-valued property "
                + getPath());
      } else {
        throw new ValueFormatException(
            "Can not assign single-value Value to a multiple-valued property "
                + getPath());
      }
    }

    // Remove this property if values == null
    if (propertyValues == null) {
      if (def.isMandatory())
        throw new ConstraintViolationException(
            "Can not remove (by setting null value) protected or mandatory property "
                + getPath());
      remove();
      return null;
    }

    // Convert to required type if neccessary
    int requiredType = def.getRequiredType();
    int propType = requiredType;
    if (propertyValues.size() > 0) {
      if (requiredType == PropertyType.UNDEFINED) {
        // requiredType = newValues[0].getType();
        if (expectedType == PropertyType.UNDEFINED) {
          expectedType = propertyValues.get(0).getType();
        }
        propType = expectedType;
      }
    }

    // Remove null values, reorder values and convert if neccessary
    List<ValueData> newValues = new ArrayList<ValueData>();
    for (int i = 0; i < propertyValues.size(); i++) {
      Value v = propertyValues.get(i);
      if (v != null) {
        ValueData valueData;
        if (v instanceof BaseValue || propType != requiredType)
          valueData = ((BaseValue) v).getData();
        else {
          BaseValue bv = (BaseValue) valueFactory.convertValue(v, propType,
              true);
          valueData = bv.getData();
        }
        ValueData newValue = valueData; 
        if(valueData.getOrderNumber() != i) {
          valueData.setOrderNumber(i);
//          if(data.isByteArray())
//            newValue = new TransientValueData(valueData.getAsByteArray(), i);
//          else
//            newValue = new TransientValueData(valueData.getAsStream(), i);
        }
        newValues.add(newValue);
      }
    }

    // Check value constraints
    checkValueConstraints(def.getValueConstraints(), newValues, propType);

    TransientPropertyData newData = TransientPropertyData.createPropertyData(
        (NodeData) parent().getData(), 
        getInternalPath().getName(), 
        propType,
        multiValue, 
        newValues);
    
    return newData;
  }

  private void checkValueConstraints(String[] valConstraints, List<ValueData> newValues, int type)
      throws ConstraintViolationException, RepositoryException {

    //String[] valConstraints = getDefinition().getValueConstraints();

    if (valConstraints != null && valConstraints.length != 0) {
      for (int j = 0; j < newValues.size(); j++) {
        // do not use getString because of string consuming
        String strVal = new String(newValues.get(j).getAsByteArray());
        boolean invalid = true;
        for (int i = 0; i < valConstraints.length; i++) {
          switch (type) {
          case PropertyType.STRING:
            if (valConstraints[i].equals(strVal))
              invalid = false;
          case PropertyType.REFERENCE:
            invalid = false;
          }
        }
        // log.warn("Value constraint is only supported for STRING property
        // (TODO) ");
        if (invalid)
          throw new ConstraintViolationException("Can not set value " + strVal
              + " to " + getPath() + " due to value constraints ");
      }
    }
  }


//  /**
//   * @author [PN] 18.04.06
//   * @see java.lang.Object#equals(java.lang.Object)
//   */
//  @Override
//  public boolean equals(Object obj) {
//    if (obj instanceof ItemImpl) {
//      ItemImpl otherItem = (ItemImpl) obj;
//      try {
//        return getInternalUUID().equals(otherItem.getInternalUUID());
//      } catch (Exception e) {
//        return false;
//      }
//    }
//    return false;
//  }
//
//  /**
//   * @see java.lang.Object#hashCode()
//   */
//  @Override
//  public int hashCode() {
//    // log.info("hashCode(): " + itemHashCode);
//    return itemHashCode;
//  }
}