/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import javax.jcr.ItemVisitor;
import javax.jcr.Node;
import javax.jcr.Property;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.PropertyDefinition;
import javax.jcr.version.VersionException;

import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.core.value.BaseValue;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: PropertyImpl.java 8214 2006-08-23 12:12:55Z peterit $
 */

public class PropertyImpl extends ItemImpl implements Property {

  private PropertyDefinition definition;

  PropertyImpl(ItemData data, SessionImpl session) throws RepositoryException, ConstraintViolationException {
    super(data, session);
    initDefinition();
  }


  public void loadData(ItemData data) throws RepositoryException {
    this.data = data;
    initDefinition();
  }

  private void initDefinition() throws RepositoryException, ConstraintViolationException {

    this.definition = parent().findChildPropertyDefinition(propertyData().getQPath().getName());
//    if (this.definition == null)
//      throw new ConstraintViolationException("Definition for property " + getPath() + " not found.");
  }

  /**
   * @see javax.jcr.Property#getValue
   */
  public Value getValue() throws ValueFormatException, RepositoryException {
    
    if (getDefinition().isMultiple())
      throw new ValueFormatException("The property " + getPath() + " is multi-valued (6.2.4)");

    // TODO BINARY re-rerad of property data (stream etc) from storage (no cache)
    
    List <ValueData> vals = propertyData().getValues();
    if (vals.isEmpty())
      throw new ValueFormatException("The single valued property " + getPath() + " is empty");
    
    // The object returned is a copy of the stored value and is immutable.
    return valueFactory.loadValue(vals.get(0), getType()); //.convertValue(values[0], this.type, true);

  }
  

  /**
   * @see javax.jcr.Property#getValues
   */
  public Value[] getValues() throws ValueFormatException, RepositoryException {
    if (!getDefinition().isMultiple() && !isMultiValued()) {
      throw new ValueFormatException("The property is single-valued (6.2.4)");
    }

    return getValueArray();
  }

  /**
   * @see javax.jcr.Property#getString
   */
  public String getString() throws ValueFormatException, RepositoryException {
    try {
      return getValue().getString();
    } catch (ValueFormatException e) {
      throw new ValueFormatException("PropertyImpl.getString() for " + getPath() + " failed: " + e);
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getString() for " + getPath() + " failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getDouble
   */
  public double getDouble() throws ValueFormatException, RepositoryException {
    try {
      return getValue().getDouble();
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getDouble() failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getLong
   */
  public long getLong() throws ValueFormatException, RepositoryException {
    try {
      return getValue().getLong();
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getLong() failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getStream
   */
  public InputStream getStream() throws ValueFormatException, RepositoryException {
    try {
      return getValue().getStream();
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getStream() failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getDate
   */
  public Calendar getDate() throws ValueFormatException, RepositoryException {
    try {
      return getValue().getDate();
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getDate() failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getBoolean
   */
  public boolean getBoolean() throws ValueFormatException, RepositoryException {
    try {
      return getValue().getBoolean();
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getBoolean() failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getNode
   */
  public Node getNode() throws ValueFormatException, RepositoryException {
    try {
      String uuid = ((BaseValue) getValue()).getReference();
      return session.getNodeByUUID(uuid);
    } catch (IllegalStateException e) {
      throw new ValueFormatException("PropertyImpl.getNode() failed: " + e);
    }
  }

  /**
   * @see javax.jcr.Property#getLength
   */
  public long getLength() throws ValueFormatException, RepositoryException {
    return lengths()[0];
  }

  /**
   * @see javax.jcr.Property#getLengths
   */
  public long[] getLengths() throws ValueFormatException, RepositoryException {
    return lengths();
  }
  
  private long[] lengths() {
    long[] lengths = new long[propertyData().getValues().size()];
    for (int i = 0; i < lengths.length; i++) {
      ValueData data = propertyData().getValues().get(i);
      lengths[i] = data.getAsByteArray().length;
    }
    return lengths;
  }

  /**
   * @see javax.jcr.Property#getDefinition
   */
  public PropertyDefinition getDefinition() throws RepositoryException {
    return this.definition;
  }

  /**
   * @see javax.jcr.Property#getType
   */
  public int getType() {
    return propertyData().getType();
  }


  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(Value value) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {

    List <Value> newValues;
    int type = PropertyType.UNDEFINED;
    if (value != null) {
      newValues = new ArrayList <Value>();
      newValues.add(value);
      type = value.getType();
    } else {
      newValues = null;
    }
    TransientPropertyData newData = doUpdateProperty(parent(), getInternalName(),
        newValues, true, type);
    
    ItemState state = ItemState.createUpdated(newData);
    dataManager.update(state, true);
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(Value[] values) throws ValueFormatException, VersionException,
      LockException, ConstraintViolationException, RepositoryException {

    List <Value> newValues = new ArrayList <Value>(values.length);
    for(Value val: values)
      newValues.add(val);
    
    int type = PropertyType.UNDEFINED;
    if (values != null) {
      if (values[0] != null)
        type = values[0].getType();
      else
        type = PropertyType.UNDEFINED;
    } else {
      newValues = null;
    }

    TransientPropertyData newData = doUpdateProperty(parent(), getInternalName(),
        newValues, true, type);
    
    ItemState state = ItemState.createUpdated(newData);
    dataManager.update(state, true);

  }
  
  /**
   * @return multiValued property of data field (PropertyData) it's a life-state
   *         property field which contains multiple-valued flag for value(s)
   *         data. Can be set in property creation time or from persistent
   *         storage.
   */
  public boolean isMultiValued() {
    return propertyData().isMultiValued();
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(String value) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {
    setValue(valueFactory.createValue(value));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(InputStream stream) throws ValueFormatException, VersionException,
      LockException, ConstraintViolationException, RepositoryException {
    setValue(valueFactory.createValue(stream));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(double number) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {
    setValue(valueFactory.createValue(number));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(long number) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {
    setValue(valueFactory.createValue(number));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(Calendar date) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {
    setValue(valueFactory.createValue(date));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(boolean b) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {
    setValue(valueFactory.createValue(b));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(Node value) throws ValueFormatException, VersionException, LockException,
      ConstraintViolationException, RepositoryException {
    // Uuid uuid = new Uuid(value.getUUID());
    setValue(valueFactory.createValue(value));
  }

  /**
   * @see javax.jcr.Property#setValue
   */
  public void setValue(String[] values) throws ValueFormatException, VersionException,
      LockException, ConstraintViolationException, RepositoryException {

    Value[] strValues = null;
    if (values != null) {
      strValues = new Value[values.length];
      for (int i = 0; i < values.length; i++)
        strValues[i] = valueFactory.createValue(values[i]);
      // new StringValue(values[i]);
    }
    setValue(strValues);
  }

  // ////////////////// Item implementation ////////////////////

  /**
   * @see javax.jcr.Item#accept
   */
  public void accept(ItemVisitor visitor) throws RepositoryException {
    visitor.visit(this);
  }

  /**
   * @see javax.jcr.Item#isNode
   */
  public boolean isNode() {
    return false;
  }

  // ////////////////////////////////

  public Value[] getValueArray() throws RepositoryException {
    
    List <ValueData> vals = propertyData().getValues();
    
    // The array returned is a copy of the stored values
    Value[] values = new Value[vals.size()];
    for(int i=0; i<vals.size(); i++) {
      ValueData data = vals.get(i);
      values[i] = valueFactory.loadValue(vals.get(0), getType());
    }
    
    return values;
  }
  
  protected PropertyData propertyData() {
    return (PropertyData) getData();
  }


  public String dump() {
    String vals = "Property " + getPath() + " values: ";
    try {  
      List <ValueData> datas = propertyData().getValues();
      for (int i = 0; i < datas.size(); i++) {
        ValueData badata = datas.get(i);
        vals += new String(badata.getAsByteArray()) + ";";
      }
    } catch(Exception e) {
      e.printStackTrace();
    }
    return vals;
  }


  //----------------------- Object.equals() ----------------------

  /**
   * @author [PN] 18.04.06
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof PropertyImpl) {
      try {
        // by path
        return getLocation().equals(((PropertyImpl) obj).getLocation());
      } catch(Exception e) {
        return false;
      }
    }
    return false;
  }  
  
  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    try {
      return getLocation().getAsString(false).hashCode();
    } catch(Exception e) {
      return super.hashCode();
    }
  }   
}