/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.WeakHashMap;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.nodetype.ConstraintViolationException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.dataflow.persistent.WorkspaceStorageDataManager;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.access.AccessManager;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.dataflow.session.SessionChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.session.TransactionableDataManager;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .<br>
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: SessionDataManager.java 8248 2006-08-24 18:48:49Z geaz $
 */
public class SessionDataManager {

  protected static Log log = ExoLogger.getLogger("jcr.SessionDataManager");

  private final SessionImpl session;

  private final SessionChangesLog changesLog;

  private final ItemPool itemsPool;

  private final SessionItemFactory itemFactory;

  private final TransactionableDataManager transactionableManager;

  private final AccessManager accessManager;

  public SessionDataManager(SessionImpl session, WorkspaceStorageDataManager workspaceDataManager)
      throws RepositoryException {
    this.session = session;

    this.transactionableManager = new TransactionableDataManager(workspaceDataManager, session.getId());
    this.changesLog = new SessionChangesLog(session.getId());
    this.itemsPool = new ItemPool();
    this.itemFactory = new SessionItemFactory(session);
    this.accessManager = session.getAccessManager();
  }

  public TransactionableDataManager getTransactManager() {
    return transactionableManager;
  }
 
  /**
   * Finds item data by internal qpath in this tnsient storage then in workspace
   * container.
   * 
   * @param path
   * @return existed item data or null if not found
   * @throws RepositoryException
   */
  public ItemData getItemData(InternalQPath path) throws RepositoryException {
//    return getItemData(path, false);
    // 1. Try in transient changes
    ItemState state = changesLog.getItemState(path);
    ItemData data = null; 
    if (state != null && !state.isDeleted()) {
      // 2. Try from txdatamanager
      data = transactionableManager.getItemData(path);
      
// [GA] i do not think we need some copy as ItemData 
// supposed to be Read-Only object
// Note: everything should be changed creating new ItemData      
      
//      if (copy && data != null && transactionableManager.txStarted())
//        // Make a copy of existed item data.
//        // Applicable for open XA transaction's data (transactionLog),
//        // a data from workspace its already a copy 
//        data = ItemDataFactory.makeItemData(data);
//    } else if (!state.isDeleted()) {
//      data = state.getData();
//      if (copy && data != null)
//        // Make a copy of existed item data.
//        // Applicable for session's level data (changesLog)
//        data = ItemDataFactory.makeItemData(data);
    }

    return data;

  }
  
//  /**
//   * Finds item data by internal qpath in this tnsient storage then in workspace
//   * container.
//   * 
//   * @param path
//   * @param copy - if true a copy of existed item data will be returned
//   * @return item or null if not found
//   * @throws RepositoryException
//   */
//  public ItemData getItemData(InternalQPath path, boolean copy) throws RepositoryException {
//
//    // 1. Try in transient changes
//    ItemState state = changesLog.getItemState(path);
//    ItemData data = null; 
//    if (state == null) {
//      // 2. Try from txdatamanager
//      data = transactionableManager.getItemData(path);
//      if (copy && data != null && transactionableManager.txStarted())
//        // Make a copy of existed item data.
//        // Applicable for open XA transaction's data (transactionLog),
//        // a data from workspace its already a copy 
//        data = ItemDataFactory.makeItemData(data);
//    } else if (!state.isDeleted()) {
//      data = state.getData();
//      if (copy && data != null)
//        // Make a copy of existed item data.
//        // Applicable for session's level data (changesLog)
//        data = ItemDataFactory.makeItemData(data);
//    }
//
//    return data;
//  }
  
  /**
   * Finds item by absolute path in this tnsient storage then in workspace
   * container.
   * 
   * @param path
   * @return in-session pooled item or null if not found
   * @throws RepositoryException
   */
//  public public ItemImpl getItem(InternalQPath path) throws RepositoryException {
//    return getItem(path, true);
//  }
  
  /**
   * Finds item by absolute path in this tnsient storage then in workspace
   * container.
   * 
   * @param path
   * @return item or null if not found
   * @throws RepositoryException
   */
  public ItemImpl getItem(InternalQPath path, boolean pool) throws RepositoryException {

    ItemData itemData = getItemData(path /*, false*/);
    
    if (itemData == null)
      return null;
    
    ItemImpl item = itemFactory.createItem(itemData);

    if (!item.hasPermission(PermissionType.READ)) {
      throw new AccessDeniedException("Access denied " + path.getAsString()
          + " for " + session.getUserID() + " (get item by path)");
    }

    if (pool)
      return itemsPool.get(item);
    
    return item;
  }

  /**
   * Finds item data by UUID in this transient storage then in workspace
   * container.
   * 
   * @param uuid 
   * @return existed item data or null if not found
   * @throws RepositoryException
   */
  public ItemData getItemDataByUUID(String uuid) throws RepositoryException {
//    return getItemDataByUUID(uuid, false);
    
    // 1. Try in transient changes
    ItemState state = changesLog.getItemState(uuid);
    ItemData data = null; 
    if (state != null && !state.isDeleted()) {
      // 2. Try from txdatamanager
      data = transactionableManager.getItemData(uuid);
    }
    
    return data;

  }
  
//[GA] i do not think we need some copy as ItemData 
//supposed to be Read-Only object
//Note: everything should be changed creating new ItemData      
//  /**
//   * Finds item data by UUID in this transient storage then in workspace
//   * container.
//   * 
//   * @param path
//   * @param copy - if true a copy of existed item data will be returned 
//   * @return item or null if not found
//   * @throws RepositoryException
//   */
//  public ItemData getItemDataByUUID(String uuid, boolean copy) throws RepositoryException {
//
//    // 1. Try in transient changes
//    ItemState state = changesLog.getItemState(uuid);
//    ItemData data = null; 
//    if (state == null) {
//      // 2. Try from txdatamanager
//      data = transactionableManager.getItemData(uuid);
//      if (copy && data != null && transactionableManager.txStarted())
//        // Make a copy of existed item data.
//        // Applicable for open XA transaction's data (transactionLog),
//        // a data from workspace its already a copy 
//        data = ItemDataFactory.makeItemData(data);
//    } else if (!state.isDeleted()) {
//      data = state.getData();
//      if (copy && data != null)
//        // Make a copy of existed item data.
//        // Applicable for session's level data (changesLog)
//        data = ItemDataFactory.makeItemData(data);
//    }
//    
//    return data;
//  }  
  
//  /**
//   * Finds item by UUID in this tnsient storage then in workspace container.
//   * 
//   * @param path
//   * @return in-session pooled item or null if not found
//   * @throws RepositoryException
//   */
//  public ItemImpl getItemByUUID(String uuid) throws RepositoryException {
//    return getItemByUUID(uuid, true);
//  }
  
  /**
   * Finds item by UUID in this tnsient storage then in workspace container.
   * 
   * @param uuid
   * @return item by UUID or null
   * @throws RepositoryException
   */
  public ItemImpl getItemByUUID(String uuid, boolean pool)
      throws RepositoryException {

    ItemData itemData = getItemDataByUUID(uuid /*, false*/);
    
    if (itemData == null)
      return null;
    
    ItemImpl item = itemFactory.createItem(itemData);

    if (!item.hasPermission(PermissionType.READ)) {
      throw new AccessDeniedException("Access denied, item with id : "
          + item.getPath() + " (get item by UUID)");
    }

    if (pool)
      return itemsPool.get(item);

    return item;
  }

// [GA] Let us not to add "util" methods to keep manager as "strong" as possible :)
// a lot amount of public methods some confuses here  
  
//  /**
//   * Return a property value data (as list) by internal qname
//   */
//  public List<ValueData> getPropertyValueData(InternalQPath qpath) throws PathNotFoundException, RepositoryException {
//
//    ItemData itemData = getTransactManager().getItemData(qpath);
//    if (itemData != null && !itemData.isNode()) {
//      PropertyData propData = (PropertyData) itemData;
//      return propData.getValues();
//    }
//    throw new PathNotFoundException("Property not found: " + qpath.getAsString());
//  }  
//  
//  /**
//   * Return a property value(s) by internal qname
//   */
//  public List<Value> getPropertyValue(InternalQPath qpath) throws PathNotFoundException, RepositoryException {
//
//    ItemData itemData = getTransactManager().getItemData(qpath);
//    if (itemData != null && !itemData.isNode()) {
//      PropertyData propData = (PropertyData) itemData;
//      List<Value> vals = new ArrayList<Value>();
//      ValueFactoryImpl vf = (ValueFactoryImpl) session.getValueFactory(); 
//      for (ValueData vd: propData.getValues()) {
//        vals.add(vf.loadValue(vd, propData.getType()));
//      }
//      return vals;
//    }
//    throw new PathNotFoundException("Property not found: " + qpath.getAsString());
//  }
  
  /**
   * TODO get from session storage as well
   * 
   * @param uuid
   * @return
   * @throws RepositoryException
   */
  List<PropertyImpl> getReferences(String uuid) throws RepositoryException {
    List<PropertyImpl> refs = new ArrayList<PropertyImpl>();
    for (PropertyData data : transactionableManager.getReferencesData(uuid)) {
      refs.add((PropertyImpl) itemFactory.createItem(data));
    }
    return refs;
  }

  /**
   * Mark the changed item as DELETED
   * 
   * @param item
   * @throws RepositoryException
   */
  public synchronized void delete(ItemState itemState) throws RepositoryException {

    List<ItemData> items = new ArrayList<ItemData>();
    traverseStoredData(itemState.getData(), items);
    for (ItemData data : items) {
      changesLog.add(ItemState.createDeleted(data));
      itemsPool.remove(data.getUUID());
    }

  }

  /**
   * Updates (adds or modifies) item state in the session transient storage
   * 
   * @param itemState -
   *          the state
   * @param pool -
   *          if true Manager force pooling this State so next calling will
   *          returna the same object Common rule: use pool = true if the Item
   *          supposed to be returned by JCR API (Node.addNode(),
   *          Node.setProperty() for ex) (NOTE: independently of pooling the
   *          Manager always return actual Item state)
   * @return
   * @throws RepositoryException
   */
  public synchronized ItemImpl update(ItemState itemState, boolean pool)
      throws RepositoryException {
    changesLog.add(itemState);
    ItemImpl item = itemFactory.createItem(itemState.getData());
    if (pool)
      return itemsPool.get(item);
    else
      return item;
  }

  /**
   * Commit changes
   * 
   * @param path
   * @throws RepositoryException
   * @throws AccessDeniedException
   * @throws ReferentialIntegrityException
   * @throws InvalidItemStateException
   */
  public synchronized void commit(InternalQPath path)
      throws RepositoryException, AccessDeniedException,
      ReferentialIntegrityException, InvalidItemStateException {

    // 1. validate all, throw an exception if validation failed
    validate(path);

    transactionableManager.save(changesLog.pullOutLog(path));
  }

  private synchronized void validate(InternalQPath path)
      throws RepositoryException, AccessDeniedException,
      ReferentialIntegrityException {

    // movedStorage.checkViolation(item.getLocation().getInternalPath());

    List<ItemState> changes = changesLog.getChanges(path);
    for (ItemState changedItem : changes) {

      // if(!changedItem.isNeedValidation())
      // continue;

      if (path.isDescendantOf(changedItem.getAncestorToSave(), false)) // /
        throw new ConstraintViolationException(
            "path is the same or descendant of either Session.move()'s destination or source node only "
                + path.getAsString());

      log.debug("SessionDataManager is VALIDATING "
          + changedItem.getData().getQPath().getAsString() + " "
          + ItemState.getStateName(changedItem.getState()) + " "
          + changedItem.getData().getUUID());

      // Check referencial integrity
      if (changedItem.isDeleted() && changedItem.isNode()) {

        log.debug("SessionDataManager is VALIDATING deleted refereceable "
            + changedItem.getData().getQPath().getAsString());

        NodeData node = (NodeData) changedItem.getData();

        // is node mix:referenceable ?
        NodeTypeManagerImpl ntm = (NodeTypeManagerImpl) session.getWorkspace()
            .getNodeTypeManager();
        if (!ntm.isNodeType(Constants.MIX_REFERENCEABLE, node
            .getMixinTypeNames()))
          return;

        List nodeRefs = getReferences(node.getUUID());
        Iterator refs = nodeRefs.iterator();

        while (refs.hasNext()) {
          PropertyData refProp = (PropertyData) refs.next();

          // if ref property is deleted in this session
          if (changesLog.getItemState(refProp.getUUID()).isDeleted())
            continue;

          NodeData refParent = (NodeData) getItemByUUID(
              refProp.getParentUUID(), false);
          AccessControlList acl = refParent.getACL();
          AccessManager am = session.getAccessManager();

          // JCRPath path =
          // session.getLocationFactory().createJCRPath(refProp.getQPath());
          if (!am.hasPermission(acl, PermissionType.READ, session.getUserID())) {
            throw new AccessDeniedException("Can not delete node "
                + node.getQPath() + " (" + node.getUUID() + ")"
                + ". It is currently the target of a REFERENCE property "
                + path.getAsString());
            // + " ("
            // + new String(((ValueData)
            // refProp.getValues().get(0)).getAsByteArray()) + ")");
          } else {
            throw new ReferentialIntegrityException("Can not delete node "
                + node.getQPath() + " (" + node.getUUID() + ")"
                + ". It is currently the target of a REFERENCE property "
                + path.getAsString());
            // + " ("
            // + new String(((ValueData)
            // refProp.getValues().get(0)).getAsByteArray()) + ")");
          }
        }
      }
    }
  }

  /**
   * Removes all pending changes of this item
   * 
   * @param item
   * @throws RepositoryException
   */
  synchronized void rollback(ItemData item) throws InvalidItemStateException,
      RepositoryException {

    changesLog.remove(item.getQPath());

  }

  /**
   * Reloads data from persistence storage
   * 
   * @param item
   * @throws RepositoryException
   */
  synchronized void refresh(ItemData item) throws InvalidItemStateException,
      RepositoryException {

    // Should we do something ?

    // looking at actual item data for the item in persistent storage
    // try {
    // if (!isValidItem(item)) {
    // throw new InvalidItemStateException("Invalid item object (removed or
    // wrong saved)");
    // }
    // } catch (InvalidItemStateException e) {
    // String message = "This Item object represents a workspace item that has
    // been removed (either by this session).";
    // throw new InvalidItemStateException(message + " Cause:[" + e.getMessage()
    // + "] Item: "
    // + item.getPath());
    // }
    //
    // // looping through changes list for items unchanged in this session
    // Iterator<ItemImpl> changes =
    // changedStorage.getChanges(item.getLocation(), false).iterator();
    //
    // while (changes.hasNext()) {
    // ItemImpl change = changes.next();
    // if (change.getState() == ItemChangeState.UNCHANGED) {
    // //change.loadData(workspaceDataManager.getItemData(change.getLocation()));
    // change.loadData(transactionableManager.getItemData(change.getLocation().getInternalPath()));
    // }
    // }
    //    
    // refreshCache(item.getLocation());
  }

  // private boolean isValidItem(ItemImpl item) throws
  // InvalidItemStateException, RepositoryException {
  //
  // // check persistent node data
  // //ItemData itemData = workspaceDataManager.getItemData(item.getLocation());
  // ItemData itemData =
  // transactionableManager.getItemData(item.getLocation().getInternalPath());
  // if (itemData == null) {
  // String message = "Item has no data in storage for item: ";
  // throw new InvalidItemStateException(message + item.getPath());
  // }
  // // check for delete state marker
  // if (item.getState() == ItemChangeState.DELETED
  // || item.getState() == ItemChangeState.AUTO_DELETED) {
  // String message = "Item has been removed or marked for removing. Item: ";
  // throw new InvalidItemStateException(message + item.getPath());
  // }
  // // check changes for this node
  // ItemImpl item2 = changedStorage.get(item.getInternalUUID());
  // if (item2 != null
  // && item.isNode() == item2.isNode()
  // && (item2.getState() == ItemChangeState.DELETED || item2.getState() ==
  // ItemChangeState.AUTO_DELETED)) {
  // String message = "This Item object represents a workspace item that has
  // been removed (either by this session). Item: ";
  // throw new InvalidItemStateException(message + item.getPath());
  // }
  // boolean result = false;
  // if (!item.isRoot()) {
  // // So, item is exists and has data, may be smth with parents
  // NodeImpl itemParent = (NodeImpl) item.getParent();
  // if (itemParent != null) {
  // // has parent
  // result = isValidItem(itemParent); // recursive
  // } // otherwise will throw RepositoryException
  // } else {
  // result = true; // roos is valid we think that
  // }
  // return result; // always true, otherwise exception will throw
  // }

  /**
   * Returns if node placed in loc or its descendants have pending changes
   * 
   * @param loc
   * @return
   */
  public boolean hasPendingChanges(InternalQPath path) {
    return (changesLog.getChanges(path).size() > 0);
  }

  public boolean isNew(String uuid) {
    if (changesLog.getItemState(uuid) == null
        || changesLog.getItemState(uuid).isDeleted())
      return false;
    List<ItemState> states = changesLog.getItemStates(uuid);
    for (ItemState state : states) {
      if (state.isAdded())
        return true;
    }
    return false;
  }

  public boolean isModified(String uuid) {
    if (changesLog.getItemState(uuid) == null
        || changesLog.getItemState(uuid).isDeleted())
      return false;
    List<ItemState> states = changesLog.getItemStates(uuid);
    for (ItemState state : states) {
      if (state.isUpdated())
        return true;
    }
    return false;
  }

  public synchronized List<NodeImpl> getChildNodes(NodeData parent, boolean pool)
      throws RepositoryException, AccessDeniedException {
    // merge data from changesLog with data from txManager
    List<NodeData> nodeDatas = transactionableManager.getChildNodesData(parent);
    List<NodeImpl> nodes = new ArrayList<NodeImpl>();

    // merge data
    for (ItemState state : changesLog.getChildStates(parent.getQPath(), true)) {
      nodeDatas.remove(state.getData());
      if (!state.isDeleted()
          && accessManager.hasPermission(parent.getACL(), PermissionType.READ,
              session.getUserID()))
        nodes.add((NodeImpl) itemFactory.createItem(state.getData()));
    }

    if (pool)
      return itemsPool.getNodes(nodes);
    else
      return nodes;

  }

  public synchronized List<PropertyImpl> getChildProperties(NodeData parent,
      boolean pool) throws RepositoryException, AccessDeniedException {

    List<PropertyData> propDatas = transactionableManager.getChildPropertiesData(parent);
    List<PropertyImpl> props = new ArrayList<PropertyImpl>();

    // merge data
    for (ItemState state : changesLog.getChildStates(parent.getQPath(), false)) {
      propDatas.remove(state.getData());
      if (!state.isDeleted()
          && accessManager.hasPermission(parent.getACL(), PermissionType.READ,
              session.getUserID()))

        props.add((PropertyImpl) itemFactory.createItem(state.getData()));
    }

    if (pool)
      return itemsPool.getProperties(props);
    else
      return props;

  }

  public synchronized List<NodeData> getChildNodesData(NodeData parent)
      throws RepositoryException {
    
    // merge data from changesLog with data from txManager
    List<NodeData> nodeDatas = transactionableManager.getChildNodesData(parent);
    List<NodeData> nodes = new ArrayList<NodeData>();

    // merge data
    for (ItemState state : changesLog.getChildStates(parent.getQPath(), true)) {
      nodeDatas.remove(state.getData());
      if (!state.isDeleted()
          && accessManager.hasPermission(parent.getACL(), PermissionType.READ,
              session.getUserID()))
        nodes.add((NodeData)state.getData());
    }

    return nodes;

  }

  public synchronized List<PropertyData> getChildPropertiesData(NodeData parent) throws RepositoryException {

    List<PropertyData> propDatas = transactionableManager
        .getChildPropertiesData(parent);
    List<PropertyData> props = new ArrayList<PropertyData>();

    // merge data
    for (ItemState state : changesLog.getChildStates(parent.getQPath(), false)) {
      propDatas.remove(state.getData());
      if (!state.isDeleted()
          && accessManager.hasPermission(parent.getACL(), PermissionType.READ,
              session.getUserID()))
        props.add((PropertyData)state.getData());
    }

    return props;

  }

  public String dump() {
    String d = "\nChanges:";
    for (ItemState item : changesLog.getAllStates()) {
      d += "\nItemImpl: isNode:" + item.isNode() + " uuid: "
          + item.getData().getUUID() + " path:" + item.getData().getQPath()
          + " state:" + ItemState.getStateName(item.getState());
    }

    // d += "\nPool:";
    // for(ItemImpl item:itemsPool.items.values())
    // d += "\nItemImpl: isNode:" + item.isNode() + " uuid: " +
    // item.getInternalUUID() + " path:"
    // + item.getPath() + " state:" + ItemState.getStateName(item.getState());
    // }

    return d;
  }

  private void traverseStoredData(ItemData parent, List<ItemData> items)
      throws RepositoryException {
    items.add(parent);
    if (parent.isNode()) {
      List<NodeData> childNodes = transactionableManager
          .getChildNodesData((NodeData) parent);
      for (NodeData childNode : childNodes) {
        traverseStoredData(childNode, items);
      }
      List<PropertyData> childProps = transactionableManager
          .getChildPropertiesData((NodeData) parent);
      for (PropertyData childProp : childProps) {
        traverseStoredData(childProp, items);
      }
    }
  }

  /**
   * Pool for touched items (stored externally or modified).
   */
  private final class ItemPool {

    // TODO WeakHashMap can't work with immutable keys! Fix it!
    private WeakHashMap<String, ItemImpl> items;

    ItemPool() {
      items = new WeakHashMap<String, ItemImpl>();
    }

    void remove(String uuid) {
      items.remove(uuid);
      // System.gc();
    }

    /**
     * @param uuid
     * @return true if item with given uuid is pooled
     * @throws RepositoryException
     */
    boolean contains(String uuid) {
      return items.containsKey(uuid);
    }
    
    /**
     * @param newItem
     * @return the item
     * @throws RepositoryException
     */
    ItemImpl get(ItemImpl newItem) throws RepositoryException {
      String uuid = newItem.getInternalUUID();
      ItemImpl item = items.get(uuid);
      if (item == null) {
        items.put(uuid, newItem);
        return newItem;
      } else {
        item.loadData(newItem.getData());
        return item;
      }
    }

    /**
     * @param nodes
     * @return child nodes
     * @throws RepositoryException
     */
    List<NodeImpl> getNodes(List<NodeImpl> nodes) throws RepositoryException {
      List<NodeImpl> children = new ArrayList<NodeImpl>();
      for (NodeImpl node : nodes) {
        String id = node.getInternalUUID();
        NodeImpl pooled = (NodeImpl) items.get(id);
        if (pooled == null) {
          items.put(id, node);
          children.add(node);
        } else {
          pooled.loadData(node.getData());
          children.add(pooled);
        }
      }
      return children;
    }

    /**
     * @param props
     * @return child properties
     * @throws RepositoryException
     */
    List<PropertyImpl> getProperties(List<PropertyImpl> props)
        throws RepositoryException {
      List<PropertyImpl> children = new ArrayList<PropertyImpl>();
      for (PropertyImpl prop : props) {
        String id = prop.getInternalUUID();
        PropertyImpl pooled = (PropertyImpl) items.get(id);
        if (pooled == null) {
          items.put(id, prop);
          children.add(prop);
        } else {
          pooled.loadData(prop.getData());
          children.add(pooled);
        }
      }
      return children;
    }

  }
}
