/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.dataflow.persistent;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import org.apache.commons.logging.Log;
import org.exoplatform.services.cache.CacheService;
import org.exoplatform.services.cache.ExoCache;
import org.exoplatform.services.jcr.config.CacheEntry;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.dataflow.persistent.ItemsPersistenceListener;
import org.exoplatform.services.jcr.dataflow.persistent.WorkspaceStorageCache;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .
 * 
 * Info: This cache implementation store item data and childs lists of item data.
 *       And it implements OBJECTS cache - i.e. returns same java object that was cached before.
 *       Same item data or list of childs will be returned from getXXX() calls. [PN] 13.04.06
 * 
 * TODO: [PN] 18.04.06 To develop rules for synchronization once again with concurent testing
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: WorkspaceStorageCacheImpl.java 8292 2006-08-26 17:37:27Z geaz $
 */

public class WorkspaceStorageCacheImpl implements WorkspaceStorageCache, ItemsPersistenceListener {

  static public int MAX_CACHE_SIZE = 200;
  static public long MAX_CACHE_LIVETIME = 600; // in sec
  
  protected static Log log = ExoLogger.getLogger("jcr.WorkspaceStorageCache");

  private ExoCache cache; 
  // for each uuid in cache(C) we will store tree path->list-of-uuids for same
  private WeakHashMap<String, List<String>> pathCache;
  
  private WeakHashMap<String, List<NodeData>> nodesCache;
  private WeakHashMap<String, List<PropertyData>> propertiesCache;

  private String   name;

  private boolean  enabled;
  
  public WorkspaceStorageCacheImpl(CacheService cacheService, WorkspaceEntry wsConfig)
      throws Exception {


    name = "jcr." + wsConfig.getUniqueName();
    cache = cacheService.getCacheInstance(name); // (WorkspaceDataCache) 
    
    //nodesCache = cacheService.getCacheInstance(name + ".nodes");
    //propertiesCache = cacheService.getCacheInstance(name + ".properties");
    //propertiesRefCache = cacheService.getCacheInstance(name + ".properties");

    CacheEntry cacheConfig = wsConfig.getCache();
    if (cacheConfig != null) {
      enabled = cacheConfig.isEnabled();
      int maxSize = Integer.parseInt(cacheConfig.getParameterValue("maxSize"));
      cache.setMaxSize(maxSize);
      pathCache = new WeakHashMap<String, List<String>>(maxSize);
      //nodesCache.setMaxSize(maxSize);
      //propertiesCache.setMaxSize(maxSize);
      nodesCache = new WeakHashMap<String, List<NodeData>>(maxSize);
      propertiesCache = new WeakHashMap<String, List<PropertyData>>(maxSize);
      //nodesRefCache = new WeakHashMap<NodeData, List<Reference<NodeData>>>(maxSize);
      //propertiesRefCache = new WeakHashMap<NodeData, List<Reference<PropertyData>>>(maxSize);
      long liveTime = Long.parseLong(cacheConfig.getParameterValue("liveTime"));
      cache.setLiveTime(liveTime);
      //nodesCache.setLiveTime(liveTime);
      //propertiesCache.setLiveTime(liveTime);
    } else {
      cache.setMaxSize(MAX_CACHE_SIZE);
      cache.setLiveTime(MAX_CACHE_LIVETIME);
      pathCache = new WeakHashMap<String, List<String>>();
      nodesCache = new WeakHashMap<String, List<NodeData>>();
      propertiesCache = new WeakHashMap<String, List<PropertyData>>();
      enabled = true;
    }
  }
  
  protected void toLog(String m) {
//    if (m.indexOf("[]:1[http://www.jcp.org/jcr/1.0]system")<0 || m.indexOf("version")>=0) {
//      log.debug(name+"." + m);
//    }
  }
  
  protected void toLog(String m, Throwable e) {
    log.error(name+"." + m + " -- " + e.getMessage());
    e.printStackTrace();
  }

  /**
   * @param uuid a UUID of item cached 
   * */
  public ItemData get(String uuid) {
    if (!enabled)
      return null;
    
    try {
      return getItem(uuid);
    } catch (Exception e) {
      return null;
    }
  }
  
  public ItemData get(InternalQPath path) {
    if (!enabled)
      return null;
    
    try {
      return getItem(path);
    } catch (Exception e) {
      return null;
    }
  }  
  
  /**
   * Called by read operations
   * 
   * @param data
   */
  public void put(ItemData data) {

    try {
      if (enabled && data != null) {
        
        // ExoCache calls synchronized already
        //synchronized (cache) {
          putItem(data);
        //}

        // add child item data to list of childs of the parent
        String parentUUID = data.getParentUUID();
        if (data.isNode()) {
          // add child node
          List<NodeData> cachedParentChilds = nodesCache.get(parentUUID);
          if (cachedParentChilds != null) {
            synchronized (cachedParentChilds) {
              int index = cachedParentChilds.indexOf(data);
              NodeData nodeData = (NodeData) data;
              if (index >= 0) {
                cachedParentChilds.set(index, nodeData);
                toLog("put()    update child node  " + nodeData.getUUID() + "  at index #" + index);
              } else {
                cachedParentChilds.add(nodeData);
                toLog("put()    add child node  " + nodeData.getUUID());
              }
            }
          }
        } else {
          // add child property
          List<PropertyData> cachedParentChilds = propertiesCache.get(parentUUID);
          if (cachedParentChilds != null) {
            synchronized (cachedParentChilds) {
              int index = cachedParentChilds.indexOf(data);
              if (index >= 0) {
                cachedParentChilds.set(index, (PropertyData) data);
                toLog("put()    update child property  " + data.getUUID() + "  at index #" + index);
              } else {
                if (cachedParentChilds.add((PropertyData) data)) {
                  toLog("put()    add child property  " + data.getUUID());
                }
              }
            }
          }
        }
      }
    } catch (Exception e) {
      toLog("Error put item data in cache: "
          + (data != null ? data.getQPath().getAsString() : "[null]"), e);
    }
  }
  
  public void addChildProperties(NodeData parentData, List<PropertyData> childItems) {
    if (enabled && parentData != null && childItems != null ) { 

      String logInfo = "parent:   " + parentData.getQPath().getAsString() + "    " + parentData.getUUID() 
        + " " + childItems.size();
      toLog("addChildProperties() >>> " + logInfo);
      
      String parentUUID = parentData.getUUID();
      String operName = ""; // for debug/trace only
      try {
        // [PN] 13.04.06
        // remove parent (no childs)
        //synchronized (cache) {
          operName = "removing parent";
          removeDeep(parentData, false);
          
          operName = "caching parent";
          putItem(parentData); // put parent in cache
         
          synchronized (propertiesCache) {
            // removing prev list of child properties from cache C
            operName = "removing child properties";
            removeChildProperties(parentUUID);
            
            operName = "caching child properties list";
            propertiesCache.put(parentUUID, childItems); // put childs in cache CP
          }
          
          operName = "caching child properties";
          putItems(childItems); // put childs in cache C
        //}
      } catch (Exception e) {
        toLog("Error in addChildProperties() " + operName + ": parent "
            + (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
      }
      toLog("addChildProperties() <<< " + logInfo);
    }        
  }
  
  public void addChildNodes(NodeData parentData, List<NodeData> childItems) {
    //  && childItems.size() > 0 - no childs it's a state too
    if (enabled && parentData != null && childItems != null ) { 

      String logInfo = "parent:   " + parentData.getQPath().getAsString() + "    " + parentData.getUUID() 
        + " " + childItems.size();
      toLog("addChildNodes() >>> " + logInfo);
      
      String parentUUID = parentData.getUUID();
      List<NodeData> removedChildNodes = null;
      String operName = ""; // for debug/trace only
      try {
        // [PN] 13.04.06
        // remove parent (no childs)
        //synchronized (cache) {
          operName = "removing parent";
          removeDeep(parentData, false);
          
          operName = "caching parent";
          putItem(parentData); // put parent in cache
         
          synchronized (nodesCache) {
            // removing prev list of child nodes from cache C
            operName = "removing child nodes";
            removedChildNodes = removeChildNodes(parentUUID, false);
            
            if (removedChildNodes != null && removedChildNodes.size()>0) {
              operName = "search for stale child nodes not contains in the new list of childs";
              List<NodeData> forRemove = new ArrayList<NodeData>();
              for (NodeData removedChildNode: removedChildNodes) {
                // used Object.equals(Object o), e.g. by UUID of nodes
                if (!childItems.contains(removedChildNode)) { 
                  // this child node has been removed from the list
                  // we should remve it recursive in C, CN, CP
                  forRemove.add(removedChildNode);
                }
              }
            
              if (forRemove.size()>0) {
                operName = "removing stale child nodes not contains in the new list of childs";
                // do remove of removed child nodes recursive in C, CN, CP
                // we need here locks on cache, nodesCache, propertiesCache 
                synchronized (propertiesCache) {
                  for (NodeData removedChildNode: forRemove) {
                    removeDeep(removedChildNode, true);
                  }
                }
              }
            }
            
            operName = "caching child nodes list";
            nodesCache.put(parentUUID, childItems); // put childs in cache CN
          }
          
          operName = "caching child nodes";
          putItems(childItems); // put childs in cache C
        //} // synchronized (cache)
      } catch (Exception e) {
        toLog("Error in addChildNodes() " + operName + ": parent "
            + (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
      }
      toLog("addChildNodes() <<< " + logInfo);
    }    
  }
  
  protected void putItem(ItemData data) throws Exception {
    String path = data.getQPath().getAsString();
    String uuid = data.getUUID();
    
//    if (data instanceof TransientItemData) {
//      toLog("TransientItemData: "+ path);
//    }
    
    toLog("putItem()    " + path + "    " + uuid + "  --  " + data);
    
    cache.put(path, data);
    cache.put(uuid, data);
    
    // Update cached items for same path but different uuids
    //updatePathTree(path, uuid);
  }
  
  protected ItemData getItem(String uuid) throws Exception {
  
    return (ItemData) cache.get(uuid);
  }
  
  /**
   * @param key a InternalQPath path of item cached
   */
  protected ItemData getItem(InternalQPath path) throws Exception {

    // ask direct cache (C)
    String myPath = path.getAsString();
    return (ItemData) cache.get(myPath);
  }  
  
  protected void putItems(List itemsList) throws Exception {
    for (Object item: itemsList) {
      putItem((ItemData) item);
    }
  }
  
  public List<NodeData> getChildNodes(NodeData parentData) {
    
    if (!enabled)
      return null;
    
    try {
      return nodesCache.get(parentData.getUUID());
    } catch (Exception e) {
      toLog("Error in getChildNodes() parentData: " + 
          (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
    }
    return null; // nothing cached
  }  
  
  public List<PropertyData> getChildProperties(NodeData parentData) {
    
    if (!enabled)
      return null;
    
    try {
      return propertiesCache.get(parentData.getUUID());      
    } catch (Exception e) {
      toLog("Error in getChildNodes() parentData: " + 
          (parentData != null ? parentData.getQPath().getAsString() : "[null]"), e);
    }
    return null; // nothing cached
  }
  
//  /**
//   * @deprecated 
//   * */
//  private void clear() {
//    if (!enabled)
//      return ;
//    
//    try {
//      toLog("<<<<<<<<<<<<<<<<<<< clear() " + cache.getCacheSize() + " >>>>>>>>>>>>>>>>>>>>>>>>");
//      cache.clearCache();
//      nodesCache.clear();
//      propertiesCache.clear();
//      //log.debug("Cache cleared >> " + cache.getCacheSize());
//    } catch (Exception e) {
//      toLog("Error clear workspace cache", e);
//      //throw new RepositoryException("Exception in clear Cache " + e);
//    }
//  }

  public boolean isEnabled() {
    return enabled;
  }

  public void setEnabled(boolean enabled) {
    this.enabled = enabled;
  }

  public void setMaxSize(int maxSize) {
    this.cache.setMaxSize(maxSize);
  }

  public void setLiveTime(long liveTime) {
    this.cache.setLiveTime(liveTime);
  }
  
  // --------------------- ItemsPersistenceListener --------------
//  public /*synchronized*/ void onBatchSaveItems(List <ItemState> itemStates, String sessionId) {
//  }

  // [PN] 18.04.06 TODO to develop more soft synchronization behaviour 
  public synchronized void onSaveItems(List <ItemState> itemStates, String sessionId) {
    
    if (!enabled)
      return ;
    
    for (Iterator i = itemStates.iterator(); i.hasNext();) {
      ItemState state = (ItemState) i.next();
      ItemData data = state.getData();
      toLog("onSaveItems() " + ItemState.getStateName(state.getState()) + " " + data.getQPath().getAsString() + " " + data.getParentUUID());

      try {
        if (state.isAdded() || state.isUpdated()) {
          if (!data.isNode() && data.getQPath().getName().equals(Constants.JCR_MIXINTYPES)) {
            //if it's a property and it's jcr:mixinTypes we have to remove parent 
            //from cache, as it out of date
            ItemData parentData = get(data.getParentUUID());
            //remove(parentData);
            removeDeep(parentData, false);
          }
          put(data);
        } else if (state.isDeleted()) {
          if (!data.isNode() && data.getQPath().getName().equals(Constants.JCR_MIXINTYPES)) {
            //if it's a property and it's jcr:mixinTypes we have to remove parent
            //from cache, as it out of date
            ItemData parentData = get(data.getParentUUID());
            //remove(parentData);
            removeDeep(parentData, false);          
          } else
            remove(data);
        }
      } catch (Exception e) {
        toLog("Error process onSaveItems action for item data: " 
            + (data != null ? data.getQPath().getAsString() : "[null]"), e);
      }
    }
  }

  // ---------------------------------------------------

  /**
   * Called by delete
   * 
   * @param data
   */
  public void remove(ItemData data) {
    if (!enabled)
      return;

    try {
      toLog("remove() " + data.getQPath().getAsString() + " " + data.getUUID());
      
      // do actual deep remove
      //synchronized (cache) {
        if (data.isNode()) {
          synchronized (propertiesCache) {
            synchronized (nodesCache) {
              removeDeep(data, true);
            }
          }
          removeSuccessors(data.getQPath().getAsString());
        } else {
          removeDeep(data, false);
        }
      //}
    } catch (Exception e) {
      toLog("Error remove item data from cache: " 
          + (data != null ? data.getQPath().getAsString() : "[null]"), e);
      //throw new RepositoryException("Exception in remove Cache " + e);
    }
  }
  
  /**
   * Deep remove of an item in all caches (C, CN, CP).   
   * Outside must be sinchronyzed by cache(C).
   * If forceDeep=true then it must be sinchronyzed by cache(CN,CP) too.
   * 
   * @param item - ItemData of item removing
   * @param forceDeep - if true then childs will be removed too, 
   *        item's parent childs (nodes or properties) will be removed also.
   *        if false - no actual deep remove will be done, 
   *        the item only and theirs 'phantom by uuid' if exists.
   * */
  protected ItemData removeDeep(ItemData item, boolean forceDeep) throws Exception {
    String myPath = item.getQPath().getAsString();
    toLog("removeDeep(" + forceDeep + ") >>> item " + myPath + " " + item.getUUID());
    
    if (forceDeep) {
      removeRelations(item);
    }
    
    cache.remove(item.getUUID());
    ItemData itemData = (ItemData) cache.remove(myPath);
    if (itemData != null && !itemData.getUUID().equals(item.getUUID())) {
      // same path but diff uuid node
      removeDeep(itemData, forceDeep);
    }
    toLog("removeDeep(" + forceDeep + ") <<< item " + myPath + " " + item.getUUID());
    return itemData;
  }
  
  /**
   * Remove item relations in the cache(C,CN,CP) by UUID.
   * Relations for a node it's a child nodes, properties and item in node's parent childs list.
   * Relations for a property it's a item in node's parent childs list. 
   * */
  protected void removeRelations(ItemData item) {
    // removing child item data from list of childs of the parent
    try {
      if (item.isNode()) {
        // removing childs of the node
        if (removeChildNodes(item.getUUID(), true) != null) {
          toLog("removeRelations() removeChildNodes() " + item.getUUID());
        }
        if (removeChildProperties(item.getUUID()) != null) {
          toLog("removeRelations() removeChildProperties() " + item.getUUID());
        }
        
        // removing child from the node's parent child nodes list
        if (removeChildNode(item.getParentUUID(), item.getUUID()) != null) {
          toLog("removeRelations() removeChildNode(parentUUID, childUUID) " 
              + item.getParentUUID() + " " + item.getUUID());
        }        
      } else {
        // removing child from the node's parent properties list
        if (removeChildProperty(item.getParentUUID(), item.getUUID()) != null) {
          toLog("removeRelations() removeChildProperty(parentUUID, childUUID) " 
              + item.getParentUUID() + " " + item.getUUID());
        }
      }
    } catch (Exception e) {
      toLog("Error in removeRelations() item: "
          + (item != null ? item.getQPath().getAsString() : "[null]"), e);
    }
  }

  protected List<NodeData> removeChildNodes(String parentUUID, boolean forceDeep) throws Exception {
    // [PN] 29.03.06 remove() instead get()
    List<NodeData> childNodes = nodesCache.remove(parentUUID);
    if (childNodes != null) {
      // we have child nodes
      for (NodeData cn: childNodes) {
        // [PN] 14.04.06 no deep here
        removeDeep(cn, forceDeep);
      }
    }
    return childNodes;
  }
  
  protected List<PropertyData> removeChildProperties(String parentUUID) throws Exception {
    // [PN] 29.03.06 remove() instead get()
    List<PropertyData> childProperties = propertiesCache.remove(parentUUID);
    if (childProperties != null) {
      // we have child properties
      for (PropertyData cp: childProperties) {
        // [PN] 13.04.06 
        removeDeep(cp, false);
      }
    }
    return childProperties;
  } 
    
  protected PropertyData removeChildProperty(String parentUUID, String childUUID) throws Exception {
    List<PropertyData> childProperties = propertiesCache.get(parentUUID);
    if (childProperties != null) {
      synchronized (childProperties) { // [PN] 14.04.06 TODO Does we need lock here? 
        int removedIndex = -1;
        for (int i=0; i<childProperties.size(); i++) {
          PropertyData cn = childProperties.get(i);
          if (cn.getUUID().equals(childUUID)) {
            removedIndex = i;
            break;
          }
        }
        if (removedIndex>=0) {
          return childProperties.remove(removedIndex);
        }
      }
    }
    return null;
  }
  
  protected NodeData removeChildNode(String parentUUID, String childUUID) throws Exception {
    List<NodeData> childNodes = nodesCache.get(parentUUID);
    if (childNodes != null) {
      synchronized (childNodes) { // [PN] 14.04.06 TODO Does we need lock here?
        int removedIndex = -1;
        for (int i=0; i<childNodes.size(); i++) {
          NodeData cn = childNodes.get(i);
          if (cn.getUUID().equals(childUUID)) {
            removedIndex = i;
            break;
          }
        }
        if (removedIndex>=0) {
          return childNodes.remove(removedIndex);
        }
      }
    }
    return null;
  }
  
//  /** Not used */
//  private NodeData getChildNode(String parentUUID, String childUUID) throws Exception {
//    List<NodeData> childNodes = nodesCache.get(parentUUID);
//    if (childNodes != null) {
//      for (NodeData cn: childNodes) {
//        if (cn.getUUID().equals(childUUID)) {
//          return cn;
//        }
//      }
//    }
//    return null;
//  }
//  
//  /** Not used */
//  private PropertyData getChildProperty(String parentUUID, String childUUID) throws Exception {
//    List<PropertyData> childProperties = propertiesCache.get(parentUUID);
//    if (childProperties != null) {
//      for (PropertyData cp: childProperties) {
//        if (cp.getUUID().equals(childUUID)) {
//          return cp;
//        }
//      }
//    }
//    return null;
//  }
  
  /**
   * Remove successors by parent path.
   * Path is a string, if an item's path starts with it then the item will be removed 
   * */
  protected void removeSuccessors(String parentPath) {
    //String parentPath = parent.getQPath().getAsString();
    if (cache instanceof HashMap) {
      Map<String, Object> cacheMap = (Map<String, Object>) cache;
      List<String> forRemove = new ArrayList<String>();
      long keysCount = 0;
      for (String key: cacheMap.keySet()) {
        if (key.startsWith(parentPath)) {
          // we should remove the entry
          forRemove.add(key);
        }
        keysCount++;
      }
      for (String key: forRemove) {
        try {
          ItemData removed = (ItemData) cache.remove(key);
          if (removed != null) {
            cache.remove(removed.getUUID());
          }
        } catch(Exception e) {
          toLog("removeSuccessors() " + parentPath + " key: " + key, e);
        }
      }
      if (forRemove.size()>0) 
        log.debug("removeSuccessors() STAT keys:" + keysCount + " removed: " + forRemove.size());
    }
  }
}
