/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.dataflow.session;

import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.RepositoryException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.dataflow.DataManager;
import org.exoplatform.services.jcr.dataflow.ItemDataChangesLog;
import org.exoplatform.services.jcr.dataflow.ItemState;
import org.exoplatform.services.jcr.dataflow.persistent.WorkspaceStorageDataManager;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.transaction.TransactionException;
import org.exoplatform.services.transaction.TransactionResource;

/**
 * Created by The eXo Platform SARL        .
 *
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov</a>
 * @version $Id: TransactionableDataManager.java 8214 2006-08-23 12:12:55Z peterit $
 */
public class TransactionableDataManager implements DataManager, TransactionResource {

  protected static Log log = ExoLogger.getLogger("jcr.TransactionableDataManager");
  
  private WorkspaceStorageDataManager storageDataManager;
  
  private String sessionId;
  
  private SessionChangesLog transactionLog;
  
  private VersionHistoryInitializer versionHistoryInitializer;
  
  public TransactionableDataManager(WorkspaceStorageDataManager storageDataManager,
      String sessionId) {
    super();
    this.storageDataManager = storageDataManager;
    this.sessionId = sessionId;
    this.versionHistoryInitializer = new VersionHistoryInitializer(this); 
  }

  // --------------- ItemDataConsumer --------

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildNodesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<NodeData> getChildNodesData(NodeData parent) throws RepositoryException {
    List <NodeData> nodes = storageDataManager.getChildNodesData(parent);

    // merge data
    if(txStarted()) {
      for(ItemState state:transactionLog.getChildStates(parent.getQPath(), true)) {
        nodes.remove(state.getData());
        if(!state.isDeleted())
          nodes.add((NodeData)state.getData());
      }
    }
    return nodes; 
  }


  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildPropertiesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<PropertyData> getChildPropertiesData(NodeData parent) throws RepositoryException {
    List <PropertyData> props = storageDataManager.getChildPropertiesData(parent);

    // merge data
    if(txStarted()) {
      for(ItemState state:transactionLog.getChildStates(parent.getQPath(), false)) {
        props.remove(state.getData());
        if(!state.isDeleted())
          props.add((PropertyData)state.getData());
      }
    }
    return props; 
  }


  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getItemData(org.exoplatform.services.jcr.datamodel.InternalQPath)
   */
  public ItemData getItemData(InternalQPath path) throws RepositoryException {
    ItemData data = null;
    if(txStarted()) {
      ItemState state = transactionLog.getItemState(path);
      if(state != null)
        data = state.getData();
    }
    if(data != null)
      return data;
    else
      return storageDataManager.getItemData(path);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getItemData(java.lang.String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException {
    ItemData data = null;
    if(txStarted()) {
      ItemState state = transactionLog.getItemState(uuid);
      if(state != null)
        data = state.getData();
    }
    if(data != null)
      return data;
    else
      return storageDataManager.getItemData(uuid);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getReferencesData(java.lang.String)
   */
  public List<PropertyData> getReferencesData(String uuid) throws RepositoryException {
    return storageDataManager.getReferencesData(uuid);
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildNodesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildNodesCount(NodeData parent) throws RepositoryException {
    if(txStarted()) {
      return getChildNodesData(parent).size();
    } else
      return storageDataManager.getChildNodesCount(parent);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildPropertiesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildPropertiesCount(NodeData parent) throws RepositoryException {
    if(txStarted()) {
      return getChildPropertiesData(parent).size();
    } else
      return storageDataManager.getChildPropertiesCount(parent);
  }
    
  // ---------------  --------

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.dataflow.TransactionResource#start(org.exoplatform.services.jcr.impl.dataflow.TransactionLifecycleManager)
   */
  public void start() {
    log.debug("tx start() "+this+" txStarted(): "+txStarted());
    if(!txStarted())
      transactionLog = new SessionChangesLog(sessionId);
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.dataflow.TransactionResource#commit(org.exoplatform.services.jcr.impl.dataflow.TransactionLifecycleManager)
   */
  public void commit() throws TransactionException {
    if(txStarted()) {
      log.debug("tx commit() "+this+"\n"+transactionLog.dump());
      try {
        storageDataManager.save(transactionLog);
        transactionLog = null;
      } catch (InvalidItemStateException e) {
        throw new TransactionException(e);
      } catch (RepositoryException e) {
        throw new TransactionException(e);
      }
    }
  }
 
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.dataflow.TransactionResource#rollback(org.exoplatform.services.jcr.impl.dataflow.TransactionLifecycleManager)
   */
  public void rollback() {
    log.debug("tx rollback() "+this+"\n"+transactionLog.dump());
    if(txStarted()) 
      transactionLog = null;
  }

  /**
   * Updates the manager with new changes
   * If transaction is started it will fill manager's changes log,
   * else just move changes to workspace storage manager
   * It saves the changes AS IS - i.e. id DOES NOT care about 
   * cloning of this objects etc. 
   * @param changes
   * @throws RepositoryException
   */
  public void save(ItemDataChangesLog changes) throws RepositoryException  {
    
    if (log.isDebugEnabled())
      log.debug("save() "+this+" txStarted(): "+txStarted());
    
    versionHistoryInitializer.saveVersionable(changes, storageDataManager.getCurrentTime());
    
    if(txStarted())
      transactionLog.addAll(changes.getAllStates());
    else
      storageDataManager.save(changes);
  }
  
  public boolean txStarted() {
    return transactionLog != null;
  }

  public WorkspaceStorageDataManager getStorageDataManager() {
    return storageDataManager;
  }
  
//  /**
//   * @deprecated
//   * Temporary method, will be removed after Session refactoring
//   * It MAKES copy of inputting data 
//   * For newly ADDED nodes it will also traverse children and add
//   * them to output(save or pass) 
//   * @param item - an item
//   * @param validate - flag indicated that item-related validation will be performed
//   */
//   public void saveItem(ItemImpl item, boolean validate) throws RepositoryException {
//
//     if (validate)
//       validate(item);
//
//     ItemDataChangesLog changeLog;
//     if (item.isNode() && item.isNew()) { // Traverse ADDED nodes and pass to
//       // add new node
//       List<ItemState> changes = new ArrayList<ItemState>();
//       traverseAddedNode(item.getActualItemData(), changes);
//       changeLog = new ItemDataChangesLog(changes, sessionId);
//     } else {
//       // add property, update, delete items
//       TransientItemData newData = copyItemData(item.getActualItemData());
//       ItemState state = new ItemState(newData, item.getState());
//       changeLog = new ItemDataChangesLog(state, sessionId);
//       //log.debug("save: " + state.getData().getQPath().getAsString() + " "
//       //   + ItemState.getStateName(state.getState()));
//     }
//     save(changeLog);
//  }
//   
// /**
//  * Temporary method, will be removed after Session refactoring
//  * It MAKES copy of inputting data 
//  * For newly ADDED nodes it will also traverse children and add
//  * them to output(save or pass) 
//  * @param item
//  */
//  public void saveItem(ItemImpl item) throws RepositoryException {
//      saveItem(item, true);
//  } 
//
//  protected void traverseAddedNode(TransientItemData item, List<ItemState> items)
//      throws RepositoryException {
// 
//    items.add(ItemState.createAdded(copyItemData(item)));
//
//    // traverse if node
//    if (item.isNode()) {
//      TransientNodeData node = (TransientNodeData)item;
//      List props = node.getChildProperties();
//      for (int i = 0; i < props.size(); i++) {
//        traverseAddedNode((TransientPropertyData) props.get(i), items);
//      }
//      List nodes = node.getChildNodes();
//      for (int i = 0; i < nodes.size(); i++) {
//        traverseAddedNode((TransientNodeData) nodes.get(i), items);
//      }
//    } 
//  }

//  private ItemDataChangesLog copyItemData(ItemDataChangesLog changes) {
//    List <ItemState> states = new ArrayList<ItemState>()
//    for(ItemState item:changes.getAllStates()) {
//      // make a copy
//      TransientItemData newData;
//
//      if (item.isNode()) {
//
//        TransientNodeData node = (TransientNodeData)item.getData();
//        newData =
//          new TransientNodeData(node.getQPath(), node.getUUID(), 
//          node.getPersistedVersion(), node.getPrimaryTypeName(), node.getMixinTypeNames(),
//          node.getOrderNumber(), node.getParentUUID(), node.getACL());
//      } else {
//        
//        TransientPropertyData prop = (TransientPropertyData)item.getData();
//        // make a copy
//        newData = new TransientPropertyData(
//            prop.getQPath(), prop.getUUID(), prop.getPersistedVersion(),
//            prop.getType(), prop.getParentUUID(), prop.isMultiValued());
//        newData.setValues(prop.getValues());
//      }
//      states.add(new ItemState(newData, item));
//    }
//    return new ItemDataChangesLog(states, sessionId);
//  }
  
  
//  public void validate(ItemImpl item) throws RepositoryException,
//      AccessDeniedException, ReferentialIntegrityException, LockException,
//      ConstraintViolationException {
//    
//    // Check permissions
//    // delete item
//    if (item.isDeleted()) {
//
//      // check for mandatory property
//      if (!item.isNode()) {
//
//        PropertyImpl prop = (PropertyImpl) item;
//        if (prop.getDefinition().isMandatory()) {
//          NodeImpl itemParent = (NodeImpl) item.getParent();
//          // Check if parent will be removed
//          if (itemParent.isRoot()
//              || itemParent.getState() != ItemState.DELETED) {
//            throw new ConstraintViolationException(
//                "Can not remove mandatory property " + prop.getPath());
//          } // else perent node will be removed too
//        }
//      }
//
//      if (!item.hasPermission(PermissionType.REMOVE))
//        throw new AccessDeniedException("Access denied: REMOVE "
//            + item.getPath()+" for: "+item.getSession().getUserID());
//
//    }
//
//    // add node
//    if (item.isNode() && item.getState() == ItemState.ADDED) {
//
//      // check for manadatory property
//      if (item.isNode()) {
//        NodeImpl node = (NodeImpl) item;
//        node.validateMandatoryChildren();
//      }
//
//      if (!item.hasPermission(PermissionType.ADD_NODE))
//        throw new AccessDeniedException("Access denied: ADD_NODE "
//            + item.getPath()+" for: "+item.getSession().getUserID());
//    }
//
//    // add/update property
//    if (!item.isNode()
//        && (item.getState() == ItemState.UPDATED || item.getState() == ItemState.ADDED)) {
//
//      if (!item.hasPermission(PermissionType.SET_PROPERTY))
//        throw new AccessDeniedException("Access denied: SET_PROPERTY "
//            + item.getPath()+" for: "+item.getSession().getUserID());
//
//    }
//  }
  
}
