/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.value.jdbc;

import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import javax.sql.DataSource;

import org.exoplatform.services.jcr.dataflow.persistent.PersistedValueData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.storage.jdbc.JDBCConnection;
import org.exoplatform.services.jcr.storage.swap.BinaryValueSwapStorage;
import org.exoplatform.services.jcr.storage.value.ValueIOChannel;
import org.exoplatform.services.jcr.util.io.LazyInputStream;

/**
 * Created by The eXo Platform SARL Author : Peter Nedonosko
 * peter.nedonosko@exoplatform.com.ua 29.08.2006
 * 
 * @version $Id: JDBCIOChannel.java 8638 2006-09-06 14:19:34Z peterit $
 */
public class JDBCIOChannel extends JDBCConnection implements ValueIOChannel {

  protected final String INSERT_VALUE_SINGLEDB              = "insert into JCR_SVALUE(DATA, ORDER_NUM, PROPERTY_ID) VALUES(?,?,?)";
  protected final String INSERT_VALUE_MULTIDB               = "insert into JCR_MVALUE(DATA, ORDER_NUM, PROPERTY_ID) VALUES(?,?,?)";

  protected final String DELETE_VALUE_SINGLEDB              = "delete from JCR_SVALUE where PROPERTY_ID=?";
  protected final String DELETE_VALUE_MULTIDB               = "delete from JCR_MVALUE where PROPERTY_ID=?";

  protected final String FIND_VALUES_BY_PROPERTYID_SINGLEDB = "delete from JCR_SVALUE where PROPERTY_ID=?";
  protected final String FIND_VALUES_BY_PROPERTYID_MULTIDB  = "delete from JCR_MVALUE where PROPERTY_ID=?";
  
  protected final String FIND_VALUE_BY_PROPERTYID_OREDERNUMB_SINGLEDB = "delete from JCR_SVALUE where PROPERTY_ID=?";
  protected final String FIND_VALUE_BY_PROPERTYID_OREDERNUMB_MULTIDB = "delete from JCR_MVALUE where PROPERTY_ID=?";

  protected final ThreadLocal<Connection> currentConnection = new ThreadLocal<Connection>();
  
  protected final BinaryValueSwapStorage binarySwap;
  
  protected class LazyJDBCInputStream extends LazyInputStream {
    
    private final String propertyId;
    private final int orderNumber;
    
    private InputStream dataStream;
    private Statement valueStatement;
    
    LazyJDBCInputStream(String propertyId, int orderNumber) {
      this.propertyId = propertyId;
      this.orderNumber = orderNumber;
    }
    
    /**
     * Return - JDBC binary stream of DATA field. 
     * NOTE: a request statement is not closes. Actual close will be performed in 
     * LazyJDBCInputStream.close() method in client code.  
     * 
     * @param propertyId - property id of desired value
     * @param orderNumber - value order number (1 - for single valued values, more othervise) 
     * @return - JDBC binary stream of DATA field 
     * @throws SQLException
     * @throws IOException
     */
    private InputStream getValueDataStream() throws SQLException {

      final PreparedStatement findValueData = getConnection().prepareStatement(FIND_VALUE_BY_PROPERTYID_OREDERNUMB);

      findValueData.setString(1, propertyId);
      findValueData.setInt(2, orderNumber);
      final ResultSet vrs = findValueData.executeQuery();
      if (vrs.next()) {
        this.valueStatement = findValueData;
        return vrs.getBinaryStream(COLUMN_DATA);
      }
      findValueData.close();
      
      // if no records - empty stream will be returned
      // TODO may be more true to use IOException("Value data not found")
      return new ByteArrayInputStream(new byte[] {});
    }
    
    protected InputStream stream() throws IllegalStateException, IOException {
      if (dataStream == null) {
        try {
          dataStream = getValueDataStream();
        } catch(SQLException e) {
          throw new IOException("Error of data stream reading (property: " 
              + propertyId + ", " + orderNumber + "), error:" + e);
        }
      }
      return dataStream;
    }
        
    @Override
    public void close() throws IOException {
      super.close(); // close stream
      if (valueStatement != null)
        try {
          // close statement
          valueStatement.close();
        } catch(SQLException e) {
          throw new IOException("Error of database statement close operation (property: " 
              + propertyId + ", " + orderNumber + "), error:" + e);
        }
    }

    @Override
    public LazyJDBCInputStream clone() {
      return new LazyJDBCInputStream(propertyId, orderNumber);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj instanceof LazyJDBCInputStream) {
        LazyJDBCInputStream conStream = (LazyJDBCInputStream) obj;
        return conStream.propertyId.equals(propertyId) && conStream.orderNumber == orderNumber;
      }
      return super.equals(obj);
    }

    @Override
    public String toString() {
      return super.toString() + ", [" + propertyId + ", " + orderNumber + "]";
    }
  }

  JDBCIOChannel(DataSource dataSource, String containerName, boolean multiDb) throws SQLException {
    super(dataSource, containerName, multiDb);
    this.binarySwap = null;
  }
  
  JDBCIOChannel(DataSource dataSource, String containerName, boolean multiDb, BinaryValueSwapStorage binarySwap) throws SQLException {
    super(dataSource, containerName, multiDb);
    this.binarySwap = binarySwap;
  }

  @Override
  protected Connection getConnection() throws SQLException {
    Connection conn = currentConnection.get();
    if (conn == null) {
      conn = dataSource.getConnection();
      currentConnection.set(conn);
    }
    return conn;
  }

  @Override
  protected void prepareQueries() throws SQLException {
    if (multiDb) {
      INSERT_VALUE = INSERT_VALUE_MULTIDB;
      DELETE_VALUE = DELETE_VALUE_MULTIDB;
      FIND_VALUES_BY_PROPERTYID = FIND_VALUES_BY_PROPERTYID_MULTIDB;
      FIND_VALUE_BY_PROPERTYID_OREDERNUMB = FIND_VALUE_BY_PROPERTYID_OREDERNUMB_MULTIDB;
    } else {
      INSERT_VALUE = INSERT_VALUE_SINGLEDB;
      DELETE_VALUE = DELETE_VALUE_SINGLEDB;
      FIND_VALUES_BY_PROPERTYID = FIND_VALUES_BY_PROPERTYID_SINGLEDB;
      FIND_VALUE_BY_PROPERTYID_OREDERNUMB = FIND_VALUE_BY_PROPERTYID_OREDERNUMB_SINGLEDB;
    }
  }

  /**
   * Nothing todo. Statements will be prepared before execution each time, as we
   * open connection by demand
   */
  @Override
  protected void prepareStatements() throws SQLException {
  }

  /**
   * Nothing todo. Close operation is not applicable as we haven't open
   * connection
   */
  public void close() {
    try {
      getConnection().close();
    } catch (SQLException e) {
      log.error("Database connection operation error: " + e, e);
    }
  }

  public void delete(String propertyId) throws IOException {
    try {
      PreparedStatement deleteValue = getConnection().prepareStatement(DELETE_VALUE);
  
      try {
        deleteValue.setString(1, propertyId);
        deleteValue.executeUpdate();
        getConnection().commit();
      } finally {
        deleteValue.close();
      }
    } catch (SQLException e) {
      try {
        getConnection().rollback();
      } catch (SQLException e1) {
        log.error("Error of delete rollback: " + e1.getMessage());
      }
      throw new IOException("Can't delete value data of property with id " + propertyId + ", error:" + e);
    }
  }

  public void write(String propertyId, List<ValueData> data) throws IOException {
    try {
      for (int i = 0; i < data.size(); i++) {
        writeValueData(propertyId, data.get(i), i);
      }
    } catch (SQLException e) {
      try {
        getConnection().rollback();
      } catch (SQLException e1) {
        log.error("Error of write rollback: " + e1.getMessage());
      }
      throw new IOException("Can't write value data of property with id " + propertyId + ", error:" + e);
    }
  }

  protected void writeValueData(String propertyId, ValueData data, int orderNumber)
      throws SQLException, IOException {

    InputStream stream = null;
    int streamLength = 0;
    if (data.isByteArray()) {
      byte[] dataBytes = data.getAsByteArray();
      stream = new ByteArrayInputStream(dataBytes);
      streamLength = dataBytes.length;
    } else {
      stream = data.getAsStream();
      if (stream instanceof FileInputStream) {
        // TODO file length
      }
      // TODO swap this stream to swap file and read file length
      streamLength = stream.available();
    }

    PreparedStatement insertValue = getConnection().prepareStatement(INSERT_VALUE);

    try {
      insertValue.setBinaryStream(1, stream, streamLength);

      // [PN] 26.04.06 Natural values order using
      data.setOrderNumber(orderNumber);
      insertValue.setInt(2, data.getOrderNumber());

      insertValue.setString(3, propertyId);
      insertValue.executeUpdate();
      getConnection().commit();
    } finally {
      insertValue.close();
    }
  }

  public List<ValueData> read(String propertyId, boolean inBuffer) throws IOException {

    List<ValueData> data = new ArrayList<ValueData>();

    try {
      
      final PreparedStatement findValue = getConnection().prepareStatement(FIND_VALUES_BY_PROPERTYID);
  
      try {
        findValue.setString(1, propertyId);
        final ResultSet valueRecords = findValue.executeQuery();
        while (valueRecords.next()) {
          final int orderNum = valueRecords.getInt(COLUMN_ORDERNUM);
          ValueData vdata = null;
          if (binarySwap != null) {
            // open JDBC stream and read all data into BLOB swap 
            // (for small data in bytes buffer inside PersistedValueData contructor) 
            InputStream jdbcStream = valueRecords.getBinaryStream(COLUMN_DATA);
            vdata = new PersistedValueData(
                jdbcStream, 
                orderNum, 
                binarySwap.generatePath(), 
                binarySwap.getMinFileSize());
          } else if (inBuffer) {
            // open JDBC stream and read all data into bytes buffer 
            InputStream jdbcStream = valueRecords.getBinaryStream(COLUMN_DATA);
            vdata = new PersistedValueData(
                jdbcStream, 
                orderNum,
                inBuffer);            
          } else {
            // if no buffer/swap enabled - using lazy readed JDBC stream
            LazyJDBCInputStream stream = new LazyJDBCInputStream(propertyId, orderNum);
            vdata = new PersistedValueData(stream, orderNum);
          }
          data.add(vdata);
        }
      } finally {
        findValue.close();
      }
      
    } catch (SQLException e) {
      throw new IOException("Can't write value data of property with id " + propertyId + ", error:" + e);
    }

    return data;
  }
}
