package org.exoplatform.services.webdav;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;

import javax.servlet.ServletContext;
import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.chain.web.servlet.ServletWebContext;
import org.apache.commons.logging.Log;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.webdav.props.PropApi;
import org.exoplatform.services.webdav.request.RequestDocApi;
import org.exoplatform.services.webdav.request.RequestDocManager;
import org.exoplatform.services.webdav.structs.DMultistatus;

public class DavContext extends ServletWebContext {
  
  private static Log log = ExoLogger.getLogger("jcr.DavContext");

  protected String _serverPrefix = null;
  protected String _serverApp = null;
  protected String _sourceWorkspace = null;
  protected String _sourcePath = null;
  protected String _destinationWorkspace = null;
  protected String _destinationPath = null;  
  
  public DavContext(ServletContext context, HttpServletRequest request, HttpServletResponse response) {
    super(context, request, response);
    
    _serverPrefix = DavPathManager.getServerPrefix(request);
    _serverApp = DavPathManager.getServletApp(request);
    _sourceWorkspace = DavPathManager.getSrcWorkspace(request);
    _sourcePath = DavPathManager.getSrcPath(request);
    
    if (request.getHeader(DavConst.Headers.DESTINATION) != null) {
      _destinationWorkspace = DavPathManager.getDestWorkspace(request);
      _destinationPath = DavPathManager.getDestPath(request);       
    }
    
  }
  
  public String getServerPrefix() {
    return _serverPrefix;
  }
  
  public String getServerApp() {
    return _serverApp;
  }
  
  public String getSrcWorkspace() {
    return _sourceWorkspace;
  }
  
  public String getSrcPath() {
    return _sourcePath;
  }
  
  public String getDestWorkspace() {
    return _destinationWorkspace;
  }
  
  public String getDestPath() {
    return _destinationPath;
  }
  
  public boolean isSameHosts() {
    return DavPathManager.isSameDestHost(request, _serverPrefix);
  }  
  
  public DavSession getSourceSession(DavSessionProvider sessionProvider, DavConfigApi configParams) {
    String authHeader = request.getHeader(DavConst.Headers.AUTHORIZATION); 
    DavSession srcSession = sessionProvider.getSession(authHeader, _sourceWorkspace);    
    if (srcSession == null) {
      return null;
    }
    
    srcSession.setConfig(configParams);
    
    srcSession.setServerPrefix(_serverPrefix);
    srcSession.setServerApp(_serverApp);

    String srcPath = _sourcePath;
    String versionName = null;
    if (request.getParameter(DavConst.DAV_VERSIONIDENTIFIER) != null) {
      versionName = request.getParameter(DavConst.DAV_VERSIONIDENTIFIER);
    } else {
      if (srcPath.indexOf("?") >= 0) {
        versionName = srcPath.substring(srcPath.indexOf("?") + 1);
        versionName = versionName.substring(DavConst.DAV_VERSIONIDENTIFIER.length() + 1);
        srcPath = srcPath.substring(0, srcPath.indexOf("?"));
      }
    }
    srcSession.setPath(srcPath);
    srcSession.setRequiredVersionName(versionName);    

    fillLockTokens(srcSession);
    return srcSession;
  }
  
  public DavSession getDestinationSession(DavSessionProvider sessionProvider) {
    String authHeader = request.getHeader(DavConst.Headers.AUTHORIZATION);
    DavSession destSession = sessionProvider.getSession(authHeader, _destinationWorkspace);
    if (destSession == null) {
      return null;
    }
    destSession.setServerPrefix(_serverPrefix);
    destSession.setPath(_destinationPath);
    
    fillLockTokens(destSession);
    return destSession;
  }
  
  protected void fillLockTokens(DavSession session) {
    String lockToken = request.getHeader(DavConst.Headers.LOCKTOKEN);
    if (lockToken != null) {      
      lockToken = lockToken.substring(1, lockToken.length() - 1);
      //System.out.println("Added locktoken - " + lockToken);
      session.addLockToken(lockToken);
    }
    
    String ifHeaderValue = request.getHeader(DavConst.Headers.IF);
    if (ifHeaderValue != null) {
      String headerLockToken = ifHeaderValue.substring(ifHeaderValue.indexOf("("));
      headerLockToken = headerLockToken.substring(2, headerLockToken.length() - 2);
      //System.out.println("Added IF locktoken - " + headerLockToken);
      session.addLockToken(headerLockToken);
    }
  }  
  
  public RequestDocApi getDocumentFromRequest() {
    return RequestDocManager.getDocumentFromRequest(request);
  }
  
  public InputStream getRequestStream() {
    try {
      return request.getInputStream();
    } catch (Exception exc) {
      log.info("Unhandled exception. " + exc.getMessage(), exc);      
    }
    return null;
  }
  
//  public byte []getRequestBody() {
//    try {
//      String contentLength = request.getHeader(DavConst.Headers.CONTENTLENGTH);
//      int length = new Integer(contentLength);
//      ByteArrayOutputStream requestBody = new ByteArrayOutputStream();
//      
//      log.info("INPUTSTREAM - " + request.getInputStream());
//      
//      int i = 0;
//      for (i = 0; i < length + 1; i++) {
//        int dataByte = request.getInputStream().read();
//        if (dataByte < 0) {
//          break;
//        }
//        requestBody.write((byte)dataByte);
//      }    
//
//      log.info("READED");
//      
//      if (i == length) {
//        return requestBody.toByteArray();
//      }      
//    } catch (Exception exc) {
//      log.info("Can't receive request body. " + exc.getMessage());
//      exc.printStackTrace();
//    }
//    return null;
//  }  

  public String getRequestHeader(String headerName) {
    return request.getHeader(headerName);
  }
  
  public void setResponseHeader(String headerName, String headerValue) {
    response.addHeader(headerName, headerValue);
  }
  
  public void answerAuthNeeded(String defAuthHeader) {
    response.addHeader(DavConst.Headers.WWWAUTHENTICATE, defAuthHeader);
    response.setStatus(DavConst.HttpStatus.AUTHNEEDED);
  }  
  
  public void answerOk() {
    response.addHeader(DavConst.Headers.CACHECONTROL, "no-cache");
    response.setStatus(DavConst.HttpStatus.OK);
  }
  
  public void answerCreated() {
    response.setStatus(DavConst.HttpStatus.CREATED);
  }  
  
  public void answerNoContent() {
    response.setStatus(DavConst.HttpStatus.NOCONTENT);
  }  
  
  public void answerNotFound() {
    response.setStatus(DavConst.HttpStatus.NOTFOUND);
  }  

  public void answerForbidden() {
    response.setStatus(DavConst.HttpStatus.FORBIDDEN);
  }  

  public void answerPreconditionFailed() {
    response.setStatus(DavConst.HttpStatus.PRECONDITIONFAILED);
  }  
  
  public void answerProperty(PropApi property) {
    try {
      DavUtil.sendSingleProperty(response, property);
    } catch (Exception exc) {
      log.info("Can't send property. " + exc.getMessage());
      exc.printStackTrace();
    }    
  }  
  
  public void answerMultistatus(DMultistatus multistatus) {
    try {
      DavUtil.SendMultistatus(response, multistatus);
    } catch (Exception exc) {
      log.info("Can't send multistatus. " + exc.getMessage());
      exc.printStackTrace();
    }
  }
  
  public void writeToResponse(byte dataByte) throws Exception {
    response.getOutputStream().write(dataByte);
  }
  
  public void writeToResponse(byte []data, int len) throws Exception {
    response.getOutputStream().write(data, 0, len);
  }
  
}
