package org.exoplatform.services.webdav;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Calendar;

import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.PathNotFoundException;
import javax.jcr.Session;
import javax.jcr.lock.Lock;
import javax.jcr.version.Version;
import javax.jcr.version.VersionHistory;
import javax.jcr.version.VersionIterator;

import org.apache.commons.logging.Log;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.webdav.props.LockDiscoveryProp;
import org.exoplatform.services.webdav.props.PropApi;
import org.exoplatform.services.webdav.props.PropManager;
import org.exoplatform.services.webdav.request.CommonPropDoc;
import org.exoplatform.services.webdav.request.PropFindDoc;
import org.exoplatform.services.webdav.request.PropertyUpdateDoc;
import org.exoplatform.services.webdav.request.RequestDocApi;
import org.exoplatform.services.webdav.request.VersionTreeDoc;
import org.exoplatform.services.webdav.structs.DavFileInfo;
import org.exoplatform.services.webdav.structs.DMultistatus;
import org.exoplatform.services.webdav.structs.DResponse;

public class DavSession {
  
  private static Log log = ExoLogger.getLogger("jcr.DavSession");

  protected Session session = null;
  protected String serverPrefix = null;
  protected String serverApp = null;
  protected String sourcePath = null;
  protected String versionName = null;
  
  protected DavConfigApi configParams = null;
  
  public DavSession(Session session) {
    this.session = session;    
  }  

  public DavSession(Session session, String workspace, String sourcePath) {
    this.session = session;
    this.sourcePath = sourcePath;
  }
  
  public void close() {
    session.logout();
  }
  
  public void addLockToken(String lockToken) {
    session.addLockToken(lockToken);
  }
  
  public void setConfig(DavConfigApi configParams) {
    this.configParams = configParams;
  }
  
  public void setServerPrefix(String serverPrefix) {
    this.serverPrefix = serverPrefix;
  }
  
  public void setServerApp(String serverApp) {
    this.serverApp = serverApp;
  }
  
  public void setPath(String sourcePath) {
    this.sourcePath = sourcePath;
  }
  
  public void setRequiredVersionName(String versionName) {
    this.versionName = versionName;
  }
  
  public String getWorkspace() {
    return session.getWorkspace().getName();
  }
  
  public String getPath() {
    return sourcePath;
  }

  protected DResponse getResponseByNode(Node node, String href, CommonPropDoc propsInfo) throws Exception {
    DResponse response = new DResponse();
    
    if (node instanceof Version) {
      response.setHref(href + DavConst.DAV_VERSIONPREFIX + node.getName());
    } else {
      response.setHref(href);
    }
    
    ArrayList<String> properties = propsInfo.getProperties();
    
    for (int i = 0; i < properties.size(); i++) {
      String propertyName = properties.get(i);
      PropApi property = PropManager.getPropertyByName(propertyName);
      property.fillFromNode(node, href);
      response.AddProperty(property);
    }
    
    return response;
  }
    
  public DMultistatus getProperties(RequestDocApi propsInfo) {
    if (!(propsInfo instanceof PropFindDoc)) {
      return null;
    }

    ArrayList<DResponse> responses = new ArrayList<DResponse>();
    
    try {
      Node resNode = null;
      
      try {
        resNode = (Node)session.getItem(sourcePath);
      } catch (PathNotFoundException pexc) {
        return null;
      }
      
      String pathNormalized = serverPrefix + "/" + session.getWorkspace().getName() + sourcePath;
      if (pathNormalized.endsWith("/")) {
        pathNormalized = pathNormalized.substring(0, pathNormalized.length() - 1);
      }

      if (resNode.isNodeType(DavConst.NodeTypes.NT_FILE)) {
        responses.add(getResponseByNode(resNode, pathNormalized, (CommonPropDoc)propsInfo));
      } else {
        responses.add(getResponseByNode(resNode, pathNormalized + "/", (CommonPropDoc)propsInfo));        
        NodeIterator nodeIter = resNode.getNodes();
        while (nodeIter.hasNext()) {
          Node curNode = nodeIter.nextNode();
          String href = pathNormalized + "/" + curNode.getName();
          if (!curNode.isNodeType(DavConst.NodeTypes.NT_FILE)) {
            href += "/"; 
          }
          responses.add(getResponseByNode(curNode, href, (CommonPropDoc)propsInfo));
        }
      }
      
    } catch (Exception exc) {
      log.info("Can't find properties. " + exc.getMessage());
      exc.printStackTrace();
      return null;
    }
    
    return new DMultistatus(responses);
  }
  
  public DMultistatus getVersions(RequestDocApi verInfo) {
    if (!(verInfo instanceof VersionTreeDoc)) {
      return null;
    }
    
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      ArrayList<DResponse> responces = new ArrayList<DResponse>();
      
      VersionHistory history = srcNode.getVersionHistory();
      VersionIterator verIter = history.getAllVersions();

      String pathNormalized = serverPrefix + "/" + session.getWorkspace().getName() + sourcePath;
      if (pathNormalized.endsWith("/")) {
        pathNormalized = pathNormalized.substring(0, pathNormalized.length() - 1);
      }

      while (verIter.hasNext()) {
        Version curVersion = verIter.nextVersion();        
        if (!curVersion.getName().equals(DavConst.NodeTypes.JCR_ROOTVERSION)) {          
          responces.add(getResponseByNode(curVersion, pathNormalized, (VersionTreeDoc)verInfo));
        }
      }
      
      return new DMultistatus(responces);
    } catch (Exception exc) {
      log.info("Can't get version information for " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    
    return null;
  }

  protected DavFileInfo getResourceHeader(Node srcNode) throws Exception {
    DavFileInfo fileInfo = new DavFileInfo();
    
    fileInfo.setName(srcNode.getName());

    if (!srcNode.isNodeType(DavConst.NodeTypes.NT_FILE)) {
      if (srcNode.hasProperty(DavConst.NodeTypes.JCR_CREATED)) {
        fileInfo.setLatstModified(srcNode.getProperty(DavConst.NodeTypes.JCR_CREATED).getString());
      }      
      fileInfo.setContentType("");
      //fileInfo.setContentType("text/html;charset=UTF-8");
    } else {
      Node contentNode = srcNode.getNode(DavConst.NodeTypes.JCR_CONTENT);
      fileInfo.setLatstModified(contentNode.getProperty(DavConst.NodeTypes.JCR_LASTMODIFIED).getString());
      fileInfo.setContentType(contentNode.getProperty(DavConst.NodeTypes.JCR_MIMETYPE).getString());
      fileInfo.setInputStream(contentNode.getProperty(DavConst.NodeTypes.JCR_DATA).getStream());
      fileInfo.setResType(false);
    }
    
    return fileInfo;    
  }
  
  public DavFileInfo getResourceHeader() {
    try {
      Node srcNode = null;
      try {
        srcNode = (Node)session.getItem(sourcePath); 
      } catch (PathNotFoundException pexc) {
        return null;
      }
      return getResourceHeader(srcNode);
    } catch (Exception exc) {
      log.info("Can't get file header. " + exc.getMessage());
      exc.printStackTrace();
    }    
    return null;
  }
  
  public boolean createCollection() {
    try {
      String []pathes = sourcePath.split("/");
      
      Node srcNode = session.getRootNode();
      for (int i = 0; i < pathes.length; i++) {
        if ("".equals(pathes[i])) {
          continue;
        }
        if (srcNode.hasNode(pathes[i])) {
          srcNode = srcNode.getNode(pathes[i]);
        } else {
          srcNode.addNode(pathes[i], configParams.getDefFolderNodeType());
          srcNode.save();
          break;
        }
      }
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't create collection " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }

  public DavFileInfo getFolder(Node resourceNode) throws Exception {
    DavFileInfo folderInfo = new DavFileInfo();
    
    ArrayList<DavFileInfo> files = new ArrayList<DavFileInfo>();
    
    NodeIterator nodeIter = resourceNode.getNodes();
    while (nodeIter.hasNext()) {
      Node curNode = nodeIter.nextNode();
      files.add(getResourceHeader(curNode));
    }
    
    HtmlContentManager htmlManager = new HtmlContentManager(serverPrefix, serverApp);
    folderInfo.setInputStream(htmlManager.getHtmlStream("/" + getWorkspace() + sourcePath, files));    
    return folderInfo;
  }
  
  public DavFileInfo getFile() {
    try {
      Node resNode = (Node)session.getItem(sourcePath);
      
      if (!resNode.isNodeType(DavConst.NodeTypes.NT_FILE)) {
        return getFolder(resNode);
      }

      if (versionName != null) {
        VersionHistory vHistory = resNode.getVersionHistory();
        Version requiredVersion = vHistory.getVersion(versionName);
        resNode = requiredVersion.getNode(DavConst.NodeTypes.JCR_FROZENNODE);
      }

      if (!resNode.hasNode(DavConst.NodeTypes.JCR_CONTENT)) {
        log.info("node has no jcr:content");
        return null;
      }
      
      Node fileNode = resNode.getNode(DavConst.NodeTypes.JCR_CONTENT);
      DavFileInfo fileInfo = new DavFileInfo();

      if (fileNode.hasProperty(DavConst.NodeTypes.JCR_MIMETYPE)) {
        fileInfo.setContentType(fileNode.getProperty(DavConst.NodeTypes.JCR_MIMETYPE).getString());
      }
      
      if (fileNode.hasProperty(DavConst.NodeTypes.JCR_LASTMODIFIED)) {
        fileInfo.setLatstModified(fileNode.getProperty(DavConst.NodeTypes.JCR_LASTMODIFIED).getString());
      }
      
      if (fileNode.hasProperty(DavConst.NodeTypes.JCR_DATA)) {
        fileInfo.setInputStream(fileNode.getProperty(DavConst.NodeTypes.JCR_DATA).getStream());
      }

      return fileInfo;
    } catch (Exception exc) {
      log.info("Can't get content for " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    
    return null;
  }  
  
  public boolean createFile(InputStream inStream) {
    try {
      String fullPath = sourcePath;
      String filePath = fullPath.substring(0, fullPath.lastIndexOf("/"));      
      if ("".equals(filePath)) {
        filePath = "/";
      }
      String fileName = fullPath.substring(fullPath.lastIndexOf("/"));
      if (fileName.startsWith("/")) {
        fileName = fileName.substring(1);
      }

      Node srcNode = (Node)session.getItem(filePath);
      
      Node fileNode = null;
      if (srcNode.hasNode(fileName)) {
        fileNode = srcNode.getNode(fileName);
        if (!fileNode.isNodeType(DavConst.NodeTypes.NT_FILE)) {
          return false;
        }
        if (fileNode.isLocked()) {
          fileNode.unlock();
        }
      } else {
        fileNode = srcNode.addNode(fileName, configParams.getDefFileNodeType());
      }
      
      Node dataNode = null;
      if (fileNode.hasNode(DavConst.NodeTypes.JCR_CONTENT)) {
        dataNode = fileNode.getNode(DavConst.NodeTypes.JCR_CONTENT);
      } else {
        dataNode = fileNode.addNode(DavConst.NodeTypes.JCR_CONTENT, DavConst.NodeTypes.NT_RESOURCE);
      }
      
      dataNode.setProperty(DavConst.NodeTypes.JCR_MIMETYPE, "text/plain");      
      dataNode.setProperty(DavConst.NodeTypes.JCR_LASTMODIFIED, session.getValueFactory().createValue(Calendar.getInstance()));
      dataNode.setProperty(DavConst.NodeTypes.JCR_DATA, inStream);
      
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't create nt:file as " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }
  
  public boolean copy(DavSession sourceSession) {
    try {
      session.getWorkspace().copy(
          sourceSession.getWorkspace(), sourceSession.getPath(), sourcePath);
      return true;
    } catch (Exception exc) {
      log.info("Can't copy resource. " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }
  
  public boolean copy(String srcWorkspace, String srcPath, String destPath) {
    try {
      session.getWorkspace().copy(srcWorkspace, srcPath, destPath);
    } catch (Exception exc) {
      log.info(String.format("Can't copy resource '%s' to '%s'", srcPath, destPath));
      exc.printStackTrace();
      return false;
    }
    return false;
  }  
  
  public boolean delete() {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);      
      srcNode.remove();
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't delete resource. " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }

  public boolean move(DavSession sourceSession) {
    try {
      if (sourceSession.getWorkspace().equals(getWorkspace())) {
        session.move(sourceSession.getPath(), sourcePath);
        session.save();
      } else {
        copy(sourceSession);
        sourceSession.delete();
      }
      return true;
    } catch (Exception exc) {
      log.info("Can't move resource. " + exc.getMessage());
      exc.printStackTrace();
    }    
    return false;
  }
  
  public LockDiscoveryProp getLockDiscovery() {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      LockDiscoveryProp lockDiscovery = new LockDiscoveryProp();
      lockDiscovery.fillFromNode(srcNode, "");
      return lockDiscovery;
    } catch (Exception exc) {
      log.info("Can't get lockDiscovery for " + sourcePath);
      exc.printStackTrace();
    }
    return null;
  }
  
  public DMultistatus tryPropPatch(RequestDocApi propUpdateDoc) {
    if (!(propUpdateDoc instanceof PropertyUpdateDoc)) {
      return null;
    }
    
    ArrayList<DResponse> responses = new ArrayList<DResponse>();
    
    DResponse response = new DResponse();    
    ArrayList<String> needsToSet = ((PropertyUpdateDoc)propUpdateDoc).getPropertiesToSet();
    for (int i = 0; i < needsToSet.size(); i++) {
      String curPropertyName = needsToSet.get(i);
      PropApi curProp = PropManager.getPropertyByName(curPropertyName);
      curProp.setStatus(DavConst.HttpStatus.CONFLICT);
      response.AddProperty(curProp);
    }
    
    ArrayList<String> needsToRemove = ((PropertyUpdateDoc)propUpdateDoc).getPropertiesToRemove();
    for (int i = 0; i < needsToRemove.size(); i++) {
      String curPropertyName = needsToRemove.get(i);
      PropApi curProp = PropManager.getPropertyByName(curPropertyName);
      curProp.setStatus(DavConst.HttpStatus.CONFLICT);
      response.AddProperty(curProp);
    }
    
    responses.add(response);
    return new DMultistatus(responses);
  }  
  
  public Lock tryLock(RequestDocApi lockInfo) {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      try {
        if (!srcNode.isNodeType(DavConst.NodeTypes.MIX_LOCKABLE)) {
          srcNode.addMixin(DavConst.NodeTypes.MIX_LOCKABLE);
          session.save();
        }        
      } catch (Exception lockExc) {
        log.info("Can't add MIX:LOCKABLE to " + sourcePath);
      }
      Lock lockResult = srcNode.lock(true, false);
      session.save();
      return lockResult;
    } catch (Exception exc) {
      log.info("Can't lock resource " + sourcePath);
      exc.printStackTrace();
    }
    return null;
  }
  
  public boolean tryUnLock() {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      srcNode.unlock();
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't unlock resource " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }
  
  public boolean setVersionable() {
    try {
      Node resNode = (Node)session.getItem(sourcePath);

      if (configParams.isAutoMixLockable()) {
        if (!resNode.isNodeType(DavConst.NodeTypes.MIX_LOCKABLE)) {
          resNode.addMixin(DavConst.NodeTypes.MIX_LOCKABLE);
        }
      }      

      if (!resNode.isNodeType(DavConst.NodeTypes.MIX_VERSIONABLE)) {
        resNode.addMixin(DavConst.NodeTypes.MIX_VERSIONABLE);
        session.save();
      }      
      
      return true;
    } catch (Exception exc) {
      log.info("Can't set versionable resource " + sourcePath);
      exc.printStackTrace();
    }
    return false;
  }
  
  public boolean tryCheckOut() {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      srcNode.checkout();
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't check-out resource " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }
  
  public boolean tryUnCheckOut() {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      srcNode.restore(srcNode.getBaseVersion(), true);
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't restore resource - " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }
  
  public boolean tryCheckIn() {
    try {
      Node srcNode = (Node)session.getItem(sourcePath);
      srcNode.checkin();
      session.save();
      return true;
    } catch (Exception exc) {
      log.info("Can't Check-In " + sourcePath + ". " + exc.getMessage());
      exc.printStackTrace();
    }
    return false;
  }
  
}
