/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.frameworks.jcr.webdav;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Enumeration;

import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.Property;
import javax.jcr.PropertyIterator;
import javax.jcr.SimpleCredentials;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.jackrabbit.server.AbstractWebdavServlet;
import org.apache.jackrabbit.webdav.DavException;
import org.apache.jackrabbit.webdav.DavLocatorFactory;
import org.apache.jackrabbit.webdav.DavMethods;
import org.apache.jackrabbit.webdav.DavResource;
import org.apache.jackrabbit.webdav.DavResourceFactory;
import org.apache.jackrabbit.webdav.DavServletResponse;
import org.apache.jackrabbit.webdav.DavSessionProvider;
import org.apache.jackrabbit.webdav.WebdavRequest;
import org.apache.jackrabbit.webdav.WebdavResponse;
import org.apache.jackrabbit.webdav.WebdavResponseImpl;
import org.apache.jackrabbit.webdav.simple.ResourceConfig;
import org.exoplatform.container.ExoContainer;
import org.exoplatform.container.ExoContainerContext;
import org.exoplatform.services.jcr.RepositoryService;
import org.exoplatform.services.jcr.impl.core.RepositoryImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.webdav.WebdavService;
import org.exoplatform.services.webdav.impl.ExoWebdavRequestImpl;

/**
 * Created by The eXo Platform SARL.
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: $
 */

public class SimpleJCRWebdavServlet extends AbstractWebdavServlet {
  private static Log log = ExoLogger.getLogger("jcr.SimpleJCRWebdavServlet");

  public static final String INIT_PARAM_RESOURCE_CONFIG = "resource-config";
  public static final String INIT_PARAM_RESOURCE_PATH_PREFIX = "resource-path-prefix";
  public static final String INIT_PARAM_AUTHENTICATE_HEADER = "authenticate-header";
  public static final String CTX_ATTR_RESOURCE_PATH_PREFIX = "jackrabbit.webdav.simple.resourcepath";
 
  private String resourcePathPrefix;
  private WebdavService davService;
  private String authenticateHeader;
  private ResourceConfig config;
  
  @Override
  public void init() throws ServletException {
      ExoContainer container = ExoContainerContext.getCurrentContainer();
    
    if(container == null)
      throw new ServletException("Current ExoContainer is not set");
    davService = (WebdavService)container.getComponentInstanceOfType(WebdavService.class);
    log.info("WebDavService = " + davService);
    log.info("Container = " + container);

    resourcePathPrefix = getInitParameter(INIT_PARAM_RESOURCE_PATH_PREFIX);
    if (resourcePathPrefix == null) {
        log.debug("Missing path prefix > setting to empty string.");
        resourcePathPrefix = "";
    } else if (resourcePathPrefix.endsWith("/")) {
        log.debug("Path prefix ends with '/' > removing trailing slash.");
        resourcePathPrefix = resourcePathPrefix.substring(0, resourcePathPrefix.length() - 1);
    }
    getServletContext().setAttribute(CTX_ATTR_RESOURCE_PATH_PREFIX, resourcePathPrefix);
    log.info(INIT_PARAM_RESOURCE_PATH_PREFIX + " = '" + resourcePathPrefix + "'");

    authenticateHeader = getInitParameter(INIT_PARAM_AUTHENTICATE_HEADER);
    if (authenticateHeader == null) {
        authenticateHeader = DEFAULT_AUTHENTICATE_HEADER;
    }
    
    String configParam = getInitParameter(INIT_PARAM_RESOURCE_CONFIG);
    if (configParam != null) {
        try {
            config = new ResourceConfig();
            config.parse(getServletContext().getResource(configParam));
        } catch (MalformedURLException e) {
            log.info("Unable to build resource filter provider.");
        }
    }        
    super.init();
  }

  @Override
  protected void service(HttpServletRequest request, HttpServletResponse response)
          throws ServletException, IOException {
  
      //String davMethod = request.getMethod();
      //log.info("DavMethod - " + davMethod);
      
      String curServletPath = request.getContextPath() + request.getServletPath();        
      String curUrl = request.getRequestURI();
      String workspaceName = curUrl.substring(curServletPath.length());
      if (workspaceName.startsWith("/")) {
          workspaceName = workspaceName.substring(1);
      }
      String []pathes = workspaceName.split("/");
      
      WebdavRequest webdavRequest = new ExoWebdavRequestImpl(request, getLocatorFactory(), pathes[0]);
      
      int methodCode = DavMethods.getMethodCode(request.getMethod());
      boolean noCache = DavMethods.isDeltaVMethod(webdavRequest) && !(DavMethods.DAV_VERSION_CONTROL == methodCode || DavMethods.DAV_REPORT == methodCode);
      WebdavResponse webdavResponse = new WebdavResponseImpl(response, noCache);
      
      try {
          if (!getDavSessionProvider().attachSession(webdavRequest)) {
              return;
          }
        
          DavResource resource = getResourceFactory().createResource(webdavRequest.getRequestLocator(), webdavRequest, webdavResponse);
          
          if (!isPreconditionValid(webdavRequest, resource)) {
              webdavResponse.sendError(DavServletResponse.SC_PRECONDITION_FAILED);
              return;
          }
          if (!execute(webdavRequest, webdavResponse, methodCode, resource)) {
              super.service(request, response);
          }
      } catch (DavException e) {
          if (e.getErrorCode() == HttpServletResponse.SC_UNAUTHORIZED) {
              webdavResponse.setHeader("WWW-Authenticate", getAuthenticateHeaderValue());
              webdavResponse.sendError(e.getErrorCode(), e.getStatusPhrase());
          } else {
              webdavResponse.sendError(e);
          }
      } finally {
          getDavSessionProvider().releaseSession(webdavRequest);
      }
      
  }  

  @Override
  protected boolean isPreconditionValid(WebdavRequest request, DavResource resource) {
    return !resource.exists() || request.matchesIfHeader(resource);
  }

  @Override
  public DavSessionProvider getDavSessionProvider() {
    return davService.getDavSessionProvider();
  }

  @Override
  public void setDavSessionProvider(DavSessionProvider provider) {
    davService.setSessionProvider(provider);
  }

  @Override
  public DavLocatorFactory getLocatorFactory() {
    return davService.getLocatorFactory(resourcePathPrefix);
  }

  @Override
  public void setLocatorFactory(DavLocatorFactory factory) {
    davService.setLocatorFactory(factory);
  }

  @Override
  public DavResourceFactory getResourceFactory() {
    return davService.getResourceFactory(config);
  }

  @Override
  public void setResourceFactory(DavResourceFactory factory) {
    davService.setResourceFactory(factory);
  }

  @Override
  public String getAuthenticateHeaderValue() {
    return authenticateHeader;
  }

  public static String getPathPrefix(ServletContext ctx) {
    return (String) ctx.getAttribute(CTX_ATTR_RESOURCE_PATH_PREFIX);
  }

}
