/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.Item;
import javax.jcr.ItemNotFoundException;
import javax.jcr.Node;
import javax.jcr.PathNotFoundException;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.ItemDefinition;
import javax.jcr.version.VersionException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.TransientItemData;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: ItemImpl.java 7271 2006-07-21 11:58:01Z peterit $
 */

public abstract class ItemImpl implements Item {

  protected static Log log = ExoLogger.getLogger("jcr.ItemImpl");
  
  protected final SessionImpl session;
  
  protected String id; 
  
  // ///////
  protected JCRPath location;

  protected String internalUUID;

  protected NodeImpl parent;

  protected int state;

  protected int persistedVersion;

  // ///////

  // used for lazy initialization nodes/properties/parent
  protected ItemData data;

  protected SessionDataManager dataManager;

  protected LocationFactory locationFactory;

  protected ValueFactoryImpl valueFactory;

  protected boolean refreshToPersistedVersion; // [PN]

  protected long accessTime; // [PN]
  
  protected int itemHashCode = 0; // [PN] 19.04.06
  
  /**
   * New constructor
   * @param id
   * @param session
   */
  protected ItemImpl(String id, SessionImpl session) {
    this.session = session;
    this.id = id;
  }  

  // new impl methods ////////////////
  
  public void resetUUID(String id) {
    this.id = id;
  }
  
  public String getParentUUID() {
    return itemData().getParentUUID();
  }
  
  public InternalQPath getInternalPath() {
    return itemData().getQPath();
  }
  
  protected ItemData itemData() {
    //return itemsSource().getData(internalUUID);
    return null;
  }
  
  ////////////////////////////////
  
  ItemImpl(ItemData data, NodeImpl parentNode, int state, SessionImpl session)
      throws RepositoryException {
    
    this.id = data.getUUID();

    this.session = session;
    this.state = state;
    this.internalUUID = data.getUUID();
    this.persistedVersion = data.getPersistedVersion();

    this.location = session.getLocationFactory().createJCRPath(data.getQPath());
    this.parent = parentNode;

    this.data = data;

    this.dataManager = session.getTransientNodesManager();
    this.locationFactory = session.getLocationFactory();
    this.valueFactory = (ValueFactoryImpl) session.getValueFactory();
    
    // [PN] 19.04.06
    itemHashCode = (session.getWorkspace().getName() + getInternalUUID()).hashCode();
    //log.info("hashCode() INIT: " + getPath() + " " + session.getWorkspace().getName() + getInternalUUID() + " --> " + itemHashCode);
  }

  void loadLocation(JCRPath location) {
    this.location = location;
  }

  /**
   * @see javax.jcr.Item#getPath
   */
  public String getPath() {
    return getLocation().getAsString(false);
  }

  /**
   * @see javax.jcr.Item#getName
   */
  public String getName() {
    return getLocation().getName().getAsString();
  }

  /**
   * @see javax.jcr.Item#getAncestor
   */
  public Item getAncestor(int degree) throws ItemNotFoundException, AccessDeniedException,
      RepositoryException {
    try {
      // 6.2.8 If depth > n is specified then an ItemNotFoundException is
      // thrown.
      if (degree < 0)
        throw new ItemNotFoundException("Workspace.getAncestor() ancestor's degree < 0.");
      int n = getDepth() - degree;
      // log.debug("Item.getAncestor(" + degree + ") depth " +
      // getDepth()+" "+n);
      if (n == 0)
        return this;
      else if (n < 0)
        throw new ItemNotFoundException(
            "Workspace.getAncestor() ancestor's degree > depth of this item.");
      else {

        JCRPath ancestorLoc = getLocation().makeAncestorPath(n);

        return findItemByPath(ancestorLoc);
      }
    } catch (PathNotFoundException e) {
      throw new ItemNotFoundException(e.getMessage(), e);
    }
  }

  /**
   * Parent is lazy initialized
   * 
   * @see javax.jcr.Item#getParent
   */
  public Node getParent() throws ItemNotFoundException, AccessDeniedException, RepositoryException {
    if (isRoot())
      throw new ItemNotFoundException("Root node does not have a parent");

    if (this.parent == null) {
      parent = (NodeImpl) findItemByPath(getLocation().makeParentPath());
    }

    if (this.parent == null) {
      reloadIfReferenceable();      
    }

    // it does not work for newly added, and not registered with dataManager
    // (for CopyVisitor for ex)
    // it is just hack - TODO remove it in new model impl !!

    // 26.01.06 Peter, tests not depends on this
    // if (parent == null) {
    // parent = (NodeImpl)
    // dataManager.getItemFromCache(this.getActualItemData().getParentUUID());
    // }

    if (parent == null) {
      throw new RepositoryException("Parent not found for " + this.getPath() + " state: "
          + getState() + " at workspace: " + session.getWorkspace().getName());
    }
    return parent;

  }

  
  /** 
   * For referenceable node item trying reload item by his internal uuid
   * For property item trying find parent by data's parent uuid and if 
   * it's referenceable will reload the property then
   * */
  boolean reloadIfReferenceable() throws RepositoryException {
    
      if (isNode()) {
        // It's a node
        NodeImpl thisNode = (NodeImpl) this;
        if (thisNode.isNodeType("mix:referenceable")) {
          // Ok, it's referenceable node. Trying retreive this node from workspace
          NodeImpl me = (NodeImpl) findItemByUUID(getInternalUUID());
          if (me != null && me != thisNode) {
            // This node live in workspace, get parent
            //log.info("===== Reloading referenceable node: " + me.getPath() + "  " + getInternalUUID());
            try {
              loadData(me.storedNodeData()); // Data from WDM
              log.info("===== Reload referenceable node DONE: " + me.getPath() + "  " + internalUUID);
              return true;
            } catch (RepositoryException e) {
              // Sorry, no parent available. Wrong state:(
            }            
          }
        }
      } else if (data != null) {
        // It's a property. Try find parent by data's uuid
        NodeImpl myParent = (NodeImpl) findItemByUUID(data.getParentUUID());
        if (myParent != null && myParent != parent && myParent.isNodeType("mix:referenceable")) {
          //log.info("===== Reloading property (" + getName() + ") with referenceable parent: " 
          //    + myParent.getPath() + "  " + myParent.getInternalUUID());
          // So, the uuid is true and if parent is referenceable- load it
          parent = myParent;
          // PathElement[] myPath = getLocation().getEntries();
          // JCRPath myNewPath = session.getLocationFactory()
          //   .createJCRPath(myParent.getLocation(), myPath[myPath.length-1].getName());
          JCRPath myLocation = session.getLocationFactory()
            .createJCRPath(myParent.getLocation(), getName());
          loadLocation(myLocation);
          // [PN] TODO Fix usage of WDM 
          //ItemData itemData = session.getTransientNodesManager().getWorkspaceDataManager().getItemData(myLocation);
          ItemData itemData = session.getTransientNodesManager().getTransactManager().getItemData(myLocation.getInternalPath());
          if (itemData != null) {
            loadData(itemData);
          }
          //log.info("===== Reload property with referenceable parent DONE: " + getPath() + "  " + getActualItemData().getUUID());
          return true;
        }
      }
      return false;    
  }
  
  /**
   * @see javax.jcr.Item#getSession
   */
  public Session getSession() {
    return session;
  }

  /**
   * @see javax.jcr.Item#getDepth
   */
  public int getDepth() {
    return getLocation().getDepth();
  }

  /**
   * @see javax.jcr.Item#isSame
   */
  public boolean isSame(Item otherItem) {
    if (otherItem == null)
      return false;

    //if (!this.getClass().getName().equals(otherItem.getClass().getName()))
    // [PN] 18.04.06 It's more precise
    if (!this.getClass().equals(otherItem.getClass())) 
      return false;

    try {
      //return this.getActualItemData().getUUID().equals(
      //    ((ItemImpl) otherItem).getActualItemData().getUUID());
      
      // [PN] 18.04.06 UUID not changes on ItemImpl
      return getInternalUUID().equals(((ItemImpl) otherItem).getInternalUUID());
    } catch (Exception e) {
      log.debug("Item.isSame() failed " + e);
      return false;
    }
  }

  /**
   * @see javax.jcr.Item#isNew
   */
  public boolean isNew() {
    return (getState() == ItemChangeState.ADDED);
  }

  /**
   * @see javax.jcr.Item#isModified
   */
  // public boolean isModified() {
  // return (getState() == ItemChangeState.UPDATED);
  // }
  // In the subclassess:
  // public boolean isNode();
  // public void accept(ItemVisitor visitor) throws RepositoryException;
  /**
   * @see javax.jcr.Item#remove
   */
  public void remove() throws RepositoryException, ConstraintViolationException, VersionException,
      LockException {

    doRemove(true);
  }


  /**
   * @author [PN] 09.04.06
   * 
   * Temporary. For internal use.
   * Remove this item on the parent node only and set state 'deleted'. 
   * Not saving the item in session or storage. 
   * 
   * @throws RepositoryException
   * @throws LockException
   */

  
  /**
   * For internal use;
   * Remove from parent and mark as DELETED 
   * @author [PN] 11.04.06
   * */
  public void removeFromParent() throws RepositoryException, LockException {
    removeFromParent(ItemChangeState.DELETED);
  }
  
  /**
   * For internal use;
   * Remove from parent and set state 
   * @author [PN] 11.04.06
   * */  
  public void removeFromParent(int setState) throws RepositoryException, LockException {
    if (isRoot())
      throw new RepositoryException("Can't remove ROOT node.");

    parent = (NodeImpl) getParent();
    
    // Check locking
    if (!parent.checkLocking())
      throw new LockException("Node " + parent.getPath() + " is locked ");

    // [2] remove from parent
    if (isNode()) {
      if (parent.removeChildNode(this) != null) {
        parent.reorderChildNodes(true); // in case updates (reordered), do it direct from storage.
      }
    } else {
      parent.removeChildProperty(this);
    }
    state = setState;
  }  
  
  public void doRemove(boolean externalValidation) throws RepositoryException,
      ConstraintViolationException, VersionException, LockException {

    if (isRoot())
      throw new RepositoryException("Can't remove ROOT node.");

    // Check constraints
    if (externalValidation) {
      ItemDefinition def;
      if (isNode())
        def = ((NodeImpl) this).getDefinition();
      else
        def = ((PropertyImpl) this).getDefinition();

      if (def.isMandatory() || def.isProtected())
        throw new ConstraintViolationException("Can't remove mandatory or protected item "
            + getPath());
    }
    // Check if versionable ancestor is not checked-in
    parent = (NodeImpl) getParent();
    if (!parent.isCheckedOut())
      throw new VersionException("Node " + parent.getPath()
          + " or its nearest ancestor is checked-in");

    // Check locking
    if (!parent.checkLocking())
      throw new LockException("Node " + parent.getPath() + " is locked ");

    // [2] remove from parent
    if (isNode()) {
      parent.removeChildNode(this);
    } else {
      parent.removeChildProperty(this);
    }

    // [1] remove from datamanager
    dataManager.delete(this);
  }

  /**
   * @see javax.jcr.Item#save
   */
  public void save() throws ReferentialIntegrityException, AccessDeniedException, LockException,
      ConstraintViolationException, InvalidItemStateException, ReferentialIntegrityException,
      RepositoryException {

    if (isNew())
      throw new RepositoryException("It is impossible to call save() on the newly added item "
          + getPath());
    session.getTransientNodesManager().commit(this);
  }
  
  /**
   * @see javax.jcr.Item#refresh
   */
  public void refresh(boolean keepChanges) throws InvalidItemStateException, RepositoryException {
    if (!keepChanges)
      session.getTransientNodesManager().rollback(this);
    else
      session.getTransientNodesManager().refresh(this);
  }

  // -------------------- Implementation ----------------------

  public abstract TransientItemData getActualItemData() throws RepositoryException;

  public String getInternalUUID() {
    return internalUUID;
  }

  public JCRPath getLocation() {
    return this.location;
  }

  public boolean isRoot() {
    return (getDepth() == 0);
  }

  public int getState() {
    return state;
  }

  /**
   * @param newState The state to set.
   */
  protected void setState(int newState) {
    // old new result
    // A -> U = A
    // D -> ? = D
    // D -> A = A
    // A -> D = U

    int origState = newState;
    if ((this.state == ItemChangeState.ADDED || this.state == ItemChangeState.AUTO_ADDED)
        && newState == ItemChangeState.UPDATED) {
      newState = state; // not change
    } else if (this.state == ItemChangeState.DELETED || this.state == ItemChangeState.AUTO_DELETED) {
      // [PN] 04.04.06, 12.04.06
      if (newState != ItemChangeState.UNCHANGED && newState != ItemChangeState.AUTO_DELETED) {
        newState = state;
      }
    } else if (this.state == ItemChangeState.ADDED && newState == ItemChangeState.DELETED) {
      newState = ItemChangeState.UNCHANGED;
    }

    this.state = newState;
    if (log.isDebugEnabled())
      log.debug("Setting state: " + ItemChangeState.getStateName(origState) 
          + " set: " + ItemChangeState.getStateName(newState) + " " + getPath());
  }

  // set state direct, no setState() usage
  @Deprecated
  protected void markUpdated() {
    state = ItemChangeState.UPDATED;
  }

  public boolean isRefreshToPersistedVersion() {
    return refreshToPersistedVersion;
  }

  public void setRefreshToPersistedVersion(boolean refreshToPersistedVersion) {
    this.refreshToPersistedVersion = refreshToPersistedVersion;
  }

  abstract void loadData(ItemData data) throws RepositoryException;

  public void setPersistedVersion(int persistedVersion) {
    this.persistedVersion = persistedVersion;
  }

  public boolean hasPermission(String action) throws RepositoryException {

    // versionStorage's children should have READ permission for all
    if (isRoot()
        || (location.isDescendantOf(locationFactory.parseAbsPath("/jcr:system/jcr:versionStorage"),
            false) && action.equals(PermissionType.READ)))
      return true;
    else if (parent == null) {
      // [PN] 16.02.06 data used instead getActualItemData(), actualy it's same.
      // Cause - a cyclic method calls on not initialized node (no childs lazy loaded)
      parent = (NodeImpl) this.dataManager.getItemByUUID(data.getParentUUID());      
    }

    if (parent == null) {
      parent = (NodeImpl) getParent();
    }
    if (parent == null) {
      throw new RepositoryException("FATAL: parent not found for " + this.getPath());
    }

    // NodeImpl parentNode = (NodeImpl) ((isRoot()) ? (Node) this :
    // (Node)this.dataManager.getItemByUUID(getActualItemData().getParentUUID()));
    boolean hasPerm = session.getAccessManager().hasPermission(
        ((NodeImpl) parent).getChildrenACL(), action, session.getUserID());
    // System.out.println("User "+session.getUserID()+" item "+getPath()+" has
    // permission "+hasPerm);
    return hasPerm;
  }


  public int getPersistedVersion() {
    return persistedVersion;
  }

  protected Item findItemByPath(JCRPath path) throws RepositoryException {
    return dataManager.getAccessibleItem(path);
  }

  protected Item findItemByUUID(String uuid) throws RepositoryException {
    return dataManager.getActualItemByUUID(uuid);
  }
  
  /**
   * @return Returns the accessTime.
   */
  public long getAccessTime() {
    return accessTime;
  }

  /**
   * @param accessTime The accessTime to set.
   */
  public void setAccessTime(long accessTime) {
    this.accessTime = accessTime;
  }  

  /**
   * @author [PN] 18.04.06
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    if (obj instanceof ItemImpl) {
      ItemImpl otherItem = (ItemImpl) obj;
      try {
        return getInternalUUID().equals(otherItem.getInternalUUID());
      } catch(Exception e) {
        return false;
      }
    }
    return false;
  }

  /**
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    //log.info("hashCode(): " + itemHashCode);
    return itemHashCode;
  }  
}