/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.util.Calendar;

import javax.jcr.Node;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.services.jcr.datamodel.Uuid;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;

/**
 * A <code>ReferenceValue</code> provides an implementation of the
 * <code>Value</code> interface representing a <code>REFERENCE</code> value
 * (a UUID of an existing node).
 * 
 * @author Stefan Guggisberg
 */
public class ReferenceValue extends BaseValue {

  public static final int TYPE = PropertyType.REFERENCE;

  private final Uuid uuid;

  /**
   * Constructs a <code>ReferenceValue</code> object representing the UUID of
   * an existing node.
   * 
   * @param target
   *          the node to be referenced
   * @throws IllegalArgumentException
   *           If <code>target</code> is nonreferenceable.
   * @throws RepositoryException
   *           If another error occurs.
   */
  ReferenceValue(Node target) throws RepositoryException {
    //super(TYPE, new TransientValueData(new Uuid(target.getUUID())));
    super(TYPE, new TransientValueData(new Uuid(target.getProperty("jcr:uuid").getString())));    
    //this.uuid = new Uuid(target.getUUID());
    this.uuid = new Uuid(target.getProperty("jcr:uuid").getString());
    //System.out.println("========== ReferenceValue " + uuid.getString() + "  target.getUUID():" + target.getUUID());
  }

  ReferenceValue(Uuid uuid) {
    super(TYPE, new TransientValueData(uuid));
    //System.out.println("SAVED VD >>>> "+uuid.getString());
    this.uuid = uuid;
  }

  ReferenceValue(ValueData data) {
    super(TYPE, data);
    //System.out.println("LOAD VD >>>> "+new String(data.getAsByteArray()));
    this.uuid = new Uuid(data.getAsByteArray());
  }

  /// temporary!
  public ReferenceValue(String uuidStr) {
    super(TYPE, new TransientValueData(uuidStr));
    this.uuid = new Uuid(uuidStr);
  }

  /**
   * Indicates whether some other object is "equal to" this one. <p/>The result
   * is <code>true</code> if and only if the argument is not <code>null</code>
   * and is a <code>ReferenceValue</code> object that represents the same
   * value as this object.
   * 
   * @param obj
   *          the reference object with which to compare.
   * @return <code>true</code> if this object is the same as the obj argument;
   *         <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof ReferenceValue) {
      ReferenceValue other = (ReferenceValue) obj;
      if (uuid == other.uuid) {
        return true;
      } else if (uuid != null && other.uuid != null) {
        return uuid.getString().equals(other.uuid.getString());
      }
    }
    return false;
  }

  /**
   * @see BaseValue#getInternalString()
   */
  protected String getInternalString() throws ValueFormatException {
    if (uuid != null) {
      return uuid.getString();
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  //----------------------------------------------------------------< Value >
  /**
   * @see Value#getDate
   */
  public Calendar getDate() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to date failed: inconvertible types");
  }

  /**
   * @see Value#getLong
   */
  public long getLong() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to long failed: inconvertible types");
  }

  /**
   * @see Value#getBoolean
   */
  public boolean getBoolean() throws ValueFormatException,
      IllegalStateException, RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to boolean failed: inconvertible types");
  }

  /**
   * @see Value#getDouble
   */
  public double getDouble() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to double failed: inconvertible types");
  }
  
  public Uuid getUuid() {
    return uuid;
  }
}