/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.dataflow.persistent;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.RepositoryException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.ItemState;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.dataflow.DataManager;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.ItemsPersistenceListener;
import org.exoplatform.services.jcr.impl.dataflow.TransientItemData;
import org.exoplatform.services.jcr.impl.dataflow.TransientPropertyData;
import org.exoplatform.services.jcr.storage.WorkspaceDataContainer;
import org.exoplatform.services.jcr.storage.WorkspaceStorageConnection;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL .<br>
 * Workspace-level data manager
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov
 *         </a>
 * @version $Id: WorkspaceDataManager.java 7583 2006-08-04 08:32:30Z peterit $
 */
public class WorkspaceDataManager implements DataManager {

  protected static Log log = ExoLogger.getLogger("jcr.WorkspaceDataManager");

  protected WorkspaceDataContainer dataContainer;

  protected List<ItemsPersistenceListener> listeners;

  protected HashMap<SessionImpl, List<ItemState>> savedItems;

  public WorkspaceDataManager(WorkspaceDataContainer dataContainer) {
    this.dataContainer = dataContainer;
    this.listeners = new ArrayList<ItemsPersistenceListener>();
    savedItems = new HashMap<SessionImpl, List<ItemState>>();
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.dataflow.DataManager#save(org.exoplatform.services.jcr.impl.dataflow.ItemDataChangesLog)
   */
  public void save(ItemDataChangesLog changesLog) throws RepositoryException {

    List<ItemState> batchItems = savedItems.get(changesLog.getSession());
    List<ItemState> deletedItems = new ArrayList<ItemState>();

    List<ItemState> changes = changesLog.getChanges();
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {

      for (ItemState itemState : changes) {
        Object itemObj = itemState.getData();
        // should be transient!
        if (itemObj instanceof TransientItemData) {
          TransientItemData data = (TransientItemData) itemObj;
          int state = itemState.getState();
          
          // [PN] For debug output
          if (log.isDebugEnabled()) {
            String path = data.getQPath().getAsString();
            if (!path.startsWith("[]:1[http://www.jcp.org/jcr/1.0]system")
                || path.startsWith("[]:1[http://www.jcp.org/jcr/1.0]system:1[http://www.jcp.org/jcr/1.0]versionStorage")
                ) {
              log.debug("["+dataContainer.getName()+"] save item: " + ItemChangeState.getStateName(state) 
                  + " " + path + " " + data.getUUID());
            }
          }
          
          data.increasePersistedVersion();
          if (state == ItemChangeState.ADDED || state == ItemChangeState.AUTO_ADDED) {
            doAdd(data, con);
          } else if (state == ItemChangeState.UPDATED) {
            doUpdate(data, con);
          } else if (state == ItemChangeState.DELETED || state == ItemChangeState.AUTO_DELETED) {
            if (batchItems != null) {
              List<ItemState> itemsForDel = new ArrayList<ItemState>();
              traverseDeletedItems(data, itemsForDel);
              if (itemsForDel.size() > 0) {
                // Don't add first item state in itemsForDel, it's already in
                for (int j = 1; j < itemsForDel.size(); j++) {
                  deletedItems.add(itemsForDel.get(j));
                }
              }
            }
            doDelete(data, con);
          }
        }
      }
      con.commit();
      con = null;
    } catch (InvalidItemStateException e) {
      throw e;
    } catch (RepositoryException e) {
      throw e;
    } catch (RuntimeException e) {
      throw e;
    } finally {
      if (con != null)
        con.rollback();
    }

    notifySaveItems(changes, changesLog.getSession());
    
    if (batchItems != null) {
      // [PN] 07.04.06
      changes.addAll(deletedItems);
      batchItems.addAll(changes);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getItemData(InternalQPath)
   */
  public ItemData getItemData(InternalQPath qpath) throws RepositoryException {
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {
      ItemData item = con.getItemData(qpath);
      return item;
    } finally {
      con.rollback();
    }
  }


  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getItemData(String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException {
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try { 
      ItemData item = con.getItemData(uuid);
      return item;
    } finally {
      con.rollback();
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getReferencesData(String)
   */
  public List<PropertyData> getReferencesData(String uuid)
      throws RepositoryException {

    // version (frozen data, successors/predecessors, labels) never included
    // TODO move to versionable WDMs
    
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {
      InternalQPath versionStorage = InternalQPath.parse(Constants.JCR_VERSION_STORAGE_URI);

      List<PropertyData> allRefs = con.getReferencesData(uuid);
      List<PropertyData> refProps = new ArrayList<PropertyData>();
      for (int i = 0; i < allRefs.size(); i++) {
        PropertyData ref = allRefs.get(i);
        // System.out.println("REF >> "+ref.getQPath().getAsString());
        if (!ref.getQPath().isDescendantOf(versionStorage, false)) {
          refProps.add(ref);
          // System.out.println("REF ADDED>> "+ref.getQPath().getAsString());
        }
      }
      return refProps;
    } finally {
      con.rollback();
    }
  }


  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildNodesData(NodeData)
   */
  public List<NodeData> getChildNodesData(NodeData nodeData) throws RepositoryException {
    
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {
      List<NodeData> childNodes = con.getChildNodesData(nodeData);
      //if (childNodes == null || childNodes.size()<=0)
      //  log.info("WDM."+dataContainer.getName()+".getChildNodesData() '" + nodeData.getQPath().getAsString() + "' - " + (childNodes != null ? childNodes.size() : "[null]"));
      return childNodes != null ? childNodes : new ArrayList<NodeData>();
    } finally {
      con.rollback();
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildPropertiesData(NodeData)
   */
  public List<PropertyData> getChildPropertiesData(NodeData nodeData)
      throws RepositoryException {
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {
      List<PropertyData> childProperties = con.getChildPropertiesData(nodeData);
      // TODO Without new ArrayList<PropertyData>() if null 
      return childProperties != null ? childProperties : new ArrayList<PropertyData>();   
    } finally {
      con.rollback();
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildPropertiesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildPropertiesCount(NodeData nodeData) throws RepositoryException {
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {
      return con.getChildPropertiesCount(nodeData);
    } finally {
      con.rollback();
    }
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.dataflow.ItemDataConsumer#getChildNodesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildNodesCount(NodeData nodeData) throws RepositoryException {
    
    WorkspaceStorageConnection con = dataContainer.openConnection();
    try {
      return con.getChildNodesCount(nodeData);
    } finally {
      con.rollback();
    }
  }

// ----------------------------------------------
  
  /**
   * Performs actual item data deleting
   * @param item to delete
   * @param con 
   * @throws RepositoryException
   * @throws InvalidItemStateException if the item is already deleted
   */
  protected void doDelete(ItemData item, WorkspaceStorageConnection con)
      throws RepositoryException, InvalidItemStateException {

    // [PN], 01.02.06, 24.03.06, 03.04.06, 05.04.06
    ItemData sameItem = con.getItemData(item.getUUID());
    if (sameItem == null) {
      // [PN] 07.05.06 TODO Refactor it!
      sameItem = con.getItemData(item.getQPath());
      if (sameItem == null) {      
        throw new InvalidItemStateException("(delete) Item "
            + item.getQPath().getAsString()
            + " not found. Probably was deleted by another session");
      }
      // [PN] 07.05.06 TODO We have to work with item from the storage, Refactor it!
      item = sameItem;
    }
    con.delete(item);
  }

  /**
   * Performs actual item data updating
   * @param item to update
   * @param con 
   * @throws RepositoryException
   * @throws InvalidItemStateException if the item not found
   * TODO compare persistedVersion number
   */
  protected void doUpdate(ItemData item, WorkspaceStorageConnection con)
      throws RepositoryException, InvalidItemStateException {

    // check once again as another session may update the same
    ItemData sameItem = con.getItemData(item.getUUID());
    if (sameItem == null) {
      // [PN] 07.05.06 TODO Refactor it!
      sameItem = con.getItemData(item.getQPath());
      if (sameItem == null) {
        throw new InvalidItemStateException("(update) Item "
            + item.getQPath().getAsString()
            + " not found. Probably was deleted by another session");
      }
      // [PN] 07.05.06 TODO We have to work with item from the storage, Refactor it!
      item = sameItem;
    }

    if (item.isNode()) {
      con.update((NodeData) item);
    } else {
      con.update((PropertyData) item);
    }
  }

  /**
   * Performs actual item data adding
   * @param item to add
   * @param con 
   * @throws RepositoryException
   * @throws InvalidItemStateException if the item is already added
   */
  protected void doAdd(ItemData item, WorkspaceStorageConnection con)
      throws RepositoryException, InvalidItemStateException {
   
    // check once again as another session may insert the same
    ItemData sameItem = con.getItemData(item.getQPath());
    if (sameItem != null && sameItem.isNode() == item.isNode()) {
      throw new InvalidItemStateException("Item "
          + item.getQPath().getAsString() + " (persisted version: "
          + sameItem.getPersistedVersion() + ") already exists in "
          + dataContainer.getName()
          + ". Probably was added by another session " + sameItem.getUUID());
    }
    // [PN] 15.06.06 Custom tunning for MySQL self-referencing FOREIGN KEY on JCR_NODE table (FOREIGN KEY DON'T WORK!)
    if (item.isNode() && item.getParentUUID() != null) {
      ItemData parentItem = con.getItemData(item.getParentUUID());
      if (parentItem == null) {
        throw new InvalidItemStateException("Parent for item "
            + item.getQPath().getAsString() + " " + item.getUUID() + " (persisted version: "
            + item.getPersistedVersion() + ") does not exists in "
            + dataContainer.getName()
            + ". Probably was deleted by another session.");
      }
    }

    if (item.isNode()) {
      con.add((NodeData) item);
    } else {
      con.add((PropertyData) item);
    }
  }

  /**
   * @return current time
   */
  public Calendar getCurrentTime() {
    return dataContainer.getCurrentTime();
  }
  

  // ---------------------------------------------

  /**
   * Adds listener to the list
   * @param listener
   */
  public void addItemPersistenceListener(ItemsPersistenceListener listener) {
    listeners.add(listener);
    log.info("Workspace Data manager of '" + this.dataContainer.getName()
        + "' registered listener: " + listener);
  }

  protected void notifySaveItems(List itemStates, SessionImpl session) {
    for (ItemsPersistenceListener listener : listeners) {
      listener.onSaveItems(itemStates, session);
    }
  }
  

  protected void traverseDeletedItems(ItemData item, List<ItemState> items)
    throws RepositoryException {

    items.add(new ItemState(item, ItemChangeState.DELETED));
    if (item.isNode()) {
      NodeData node = (NodeData) item;
      List<PropertyData> props = getChildPropertiesData(node);
      for (PropertyData childProp : props) {
        traverseDeletedItems(childProp, items);
      }
      List<NodeData> nodes = getChildNodesData(node);
      for (NodeData childNode : nodes) {
        // [PN] 05.04.06 check for duplicate state
        // we need that for a case when children has already
        // been removed before in the current session
        if (!containsState(childNode.getQPath(), ItemChangeState.DELETED, items)) {
          traverseDeletedItems(childNode, items);
        }
      }
    }
  }

  protected boolean containsState(InternalQPath qPath, int state, List<ItemState> states) {
    if (states == null)
      return false;

    for (ItemState aState : states) {
      if (aState.getData().getQPath().equals(qPath)
          && aState.getState() == state) {
        return true;
      }
    }
    return false;
  }

}