package org.exoplatform.services.webdav.impl;

import javax.jcr.Credentials;
import javax.jcr.LoginException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.SimpleCredentials;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.codec.binary.Base64;
import org.apache.jackrabbit.server.CredentialsProvider;
import org.apache.jackrabbit.server.SessionProviderImpl;
import org.apache.jackrabbit.webdav.DavConstants;
import org.apache.jackrabbit.webdav.DavException;
import org.apache.jackrabbit.webdav.DavServletResponse;
import org.apache.jackrabbit.webdav.DavSession;
import org.apache.jackrabbit.webdav.DavSessionProvider;
import org.apache.jackrabbit.webdav.WebdavRequest;
import org.apache.jackrabbit.webdav.jcr.JcrDavException;
import org.apache.jackrabbit.webdav.simple.DavSessionImpl;
import org.exoplatform.container.xml.InitParams;
import org.exoplatform.container.xml.PropertiesParam;
import org.exoplatform.services.jcr.RepositoryService;
import org.exoplatform.services.jcr.core.ManageableRepository;

public class ExoDavSessionProvider implements DavSessionProvider {
  
  /** the 'missing-auth-mapping' init parameter */
  public final static String INIT_PARAM_MISSING_AUTH_MAPPING = "missing-auth-mapping";  
  //public final static String WORKSPACE = "workspace";  

  private ManageableRepository repository;
  private String missingAuthMapping;

  private SessionProviderImpl sessionProvider;
  //private String workspace;

  public ExoDavSessionProvider(InitParams params, 
      RepositoryService repositoryService) throws Exception {
    PropertiesParam props = params.getPropertiesParam("session.params");
    repository = repositoryService.getRepository();
    
    missingAuthMapping = props.getProperty(INIT_PARAM_MISSING_AUTH_MAPPING);
    //workspace = props.getProperty(WORKSPACE);

    CredentialsProvider cp = new CredentialsProvider() {
      public Credentials getCredentials(HttpServletRequest request) throws LoginException, ServletException {
          return getCredentialsFromHeader(request.getHeader(DavConstants.HEADER_AUTHORIZATION));
      }
    };

    sessionProvider = new SessionProviderImpl(cp);
  }
  
  
  /**
   * Acquires a DavSession. Upon success, the WebdavRequest will reference that
   * session. A session will not be available if an exception is thrown.
   * 
   * @param request
   * @throws DavException
   *           if a problem occurred while obtaining the session
   * @see DavSessionProvider#acquireSession(org.apache.jackrabbit.webdav.WebdavRequest)
   */
  
  public boolean attachSession(WebdavRequest request) throws DavException {
    try {
      // login to repository and workspace null
      String workSpaceName = ((ExoWebdavRequestImpl)request).GetWorkspaceName();
      
      Session repSession = sessionProvider.getSession(request, repository, workSpaceName);        
      if (repSession == null) {
        return false;
      }      
      DavSession ds = new DavSessionImpl(repSession);
      request.setDavSession(ds);
      return true;
    } catch (LoginException e) {
        throw new DavException(HttpServletResponse.SC_UNAUTHORIZED);
    } catch (RepositoryException e) {
        throw new JcrDavException(e);
    } catch (ServletException e) {
      throw new DavException(DavServletResponse.SC_INTERNAL_SERVER_ERROR, e
          .getMessage());
    }
  }

  /**
   * Only removes the <code>DavSession</code> object from the given request
   * object. No further actions required, since <code>DavSessionImpl</code>
   * does not allow to keep track of references to it.
   * 
   * @param request
   * @see DavSessionProvider#releaseSession(org.apache.jackrabbit.webdav.WebdavRequest)
   */
  public void releaseSession(WebdavRequest request) {
    request.setDavSession(null);
  }

  /**
   * Build a {@link Credentials} object for the given authorization header. The
   * creds may be used to login to the repository. If the specified header
   * string is <code>null</code> or not of the required format the behaviour
   * depends on the {@link #INIT_PARAM_MISSING_AUTH_MAPPING} param:<br>
   * <ul>
   * <li> if this init-param is missing, a LoginException is thrown. This is
   * suiteable for clients (eg. webdav clients) for with sending a proper
   * authorization header is not possible, if the server never send a 401.
   * <li> if this init-param is present, but with an empty value,
   * null-credentials are returned, thus forcing an null login on the repository
   * <li> if this init-param has a 'user:password' value, the respective simple
   * credentials are generated.
   * </ul>
   * 
   * @param authHeader
   *          Authorization header as present in the Http request
   * @return credentials or <code>null</code>.
   * @throws ServletException
   *           If an IOException occured while decoding the Authorization
   *           header.
   * @throws LoginException
   *           if no suitable auth header and missing-auth-mapping is not
   *           present
   * @see #getRepository()
   * @see #login(HttpServletRequest)
   */
  
  public Credentials getCredentialsFromHeader(String authHeader) 
  throws ServletException, LoginException {  
      try {
          if (authHeader != null) {
              String []authStr = authHeader.split(" ");
              
              if (authStr.length >= 2 && authStr[0].equalsIgnoreCase(HttpServletRequest.BASIC_AUTH)) {
                  String decAuthStr = new String(Base64.decodeBase64(authStr[1].getBytes()));                  
                  int pos = decAuthStr.indexOf(":");
                  String userId = decAuthStr.substring(0, pos);
                  String userPass = decAuthStr.substring(pos + 1);                  
                  return new SimpleCredentials(userId, userPass.toCharArray()); 
              }               
          }
    
          if (missingAuthMapping != null) {
              if (missingAuthMapping.equals("")) {
                  throw new LoginException();
              }
              int pos = missingAuthMapping.indexOf(':');
              if (pos < 0) {
                  return new SimpleCredentials(missingAuthMapping, null);
              }
              return new SimpleCredentials(missingAuthMapping.substring(0, pos), missingAuthMapping.substring(pos + 1).toCharArray());                   
          }
      } catch (Exception exc) {
          throw new ServletException("Unable to decode authorization: " + exc.toString());
      }       
      //throw new LoginException();
      throw new LoginException();
    }

}